#define _ACTIVITY_C
#include <Python.h>
#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
#include <numpy/arrayobject.h>

#include "_Activity.h"
#include "pysplineaddon.h"
#include "pyarraymacros.h"
#include "_StringList.h"
#include "SUNDIALSflags55.h"

#define CONSTANT 0
#define PIECEWISE_CONSTANT 1
#define PIECEWISE_LINEAR 2
#define CUBIC_SPLINE 3
#define CUBIC_SPLINE_SLOPE 4

/*
========================================================================================================================
Helper functions
========================================================================================================================
*/

/*
==========================================================================================
Output functions
==========================================================================================
*/
static double Constant_function(ConstantOutput *self, double time_local){
    return self->constant;
}

static double PiecewiseConstant_function(PiecewiseOutput *self, double time_local){
	return piecewise(PYSIZE(self->tvalues), time_local, PYDATA(self->tvalues), PYDATA(self->fvalues));
}

static double PiecewiseLinear_function(PiecewiseOutput *self, double time_local){
    return linearInterpolate(PYSIZE(self->tvalues), time_local, PYDATA(self->tvalues), PYDATA(self->fvalues));
}

static double CubicInterpolate_function(CubicSplineOutput *self, double time_local){
    return seval(PYSIZE(self->tvalues), time_local, PYDATA(self->tvalues), PYDATA(self->fvalues), 
                                        self->b, self->c, self->d, &self->last);
}

static double CubicInterpolateSlope_function(CubicSplineOutput *self, double time_local){
    return deriv(PYSIZE(self->tvalues), time_local, PYDATA(self->tvalues), 
                        self->b, self->c, self->d, &self->last);
}

/*
==========================================================================================
GETTERS AND SETTERS
==========================================================================================
*/
static PyObject *
Activity_GetOutputs(ActivityObject *self, void *closure)
{
    Py_INCREF(self->outputnames);
    return self->outputnames;
}

static int
Activity_SetOutputs(ActivityObject *self, PyObject *value, void *closure)
{
    return StringList_Update(self->outputnames, value);
}

static PyObject *
Activity_GetFeatures(ActivityObject *self, void *closure)
{
    Py_INCREF(self->featurenames);
    return self->featurenames;
}

static int
Activity_SetFeatures(ActivityObject *self, PyObject *value, void *closure)
{
    return StringList_Update(self->featurenames, value);
}

static PyObject *
Activity_GetFeatureunits(ActivityObject *self, void *closure)
{
    Py_INCREF(self->featureunits);
    return self->featureunits;
}

static int
Activity_SetFeatureunits(ActivityObject *self, PyObject *value, void *closure)
{
    return StringList_Update(self->featureunits, value);
}

static PyObject *
Activity_GetTimeUnit(ActivityObject *self, void *closure)
{
    Py_INCREF(self->timeunit);
    return self->timeunit;
}

static PyObject *
Activity_GetCompartment(ActivityObject *self, void *closure)
{
    Py_INCREF(self->compartment);
    return self->compartment;
}

static PyGetSetDef Activity_getsetters[] = {
    {"outputnames", (getter) Activity_GetOutputs, (setter) Activity_SetOutputs,
     "Output names", NULL},
     {"featurenames", (getter) Activity_GetFeatures, (setter) Activity_SetFeatures,
     "Feature names", NULL},
     {"featureunits", (getter) Activity_GetFeatureunits, (setter) Activity_SetFeatureunits,
     "Feature units", NULL},
     {"timeunit", (getter) Activity_GetTimeUnit, NULL,
     "Activity time unit", NULL},
     {"compartment", (getter) Activity_GetCompartment, NULL,
     "Activity compartment", NULL},
    {NULL}  /* Sentinel */
};

/*
==========================================================================================
PyMethods
==========================================================================================
*/
static void
Activity_AddFeature(ActivityObject *self, ActivityOutput *activity, PyObject *name, PyObject *unit){
    PyObject *tmp;

    PyList_Append(self->featurenames, name);
    if(unit)
        PyList_Append(self->featureunits, unit);
    else{
        tmp = PyUnicode_FromString("1");
        PyList_Append(self->featureunits, tmp);
        Py_DECREF(tmp);
    }

    self->nrfeatures++;
    self->features = PyMem_Realloc(self->features, self->nrfeatures * sizeof(ActivityOutput *));
    self->features[self->nroutputs-1] = activity;
}

// allocation functions
static PyObject *
Activity_AddOutput(ActivityObject *self, PyObject *args, PyObject *kwds){
    ActivityOutput *newOutput;
    static char *kwlist[] = {"type", "name", "tvalues", "fvalues", "slopeendpoints", "unit", "feature", NULL};
    PyObject *name = NULL, *tvalues = NULL, *fvalues = NULL, *slopeendpoints = NULL, *unit = NULL, *feature = NULL;
    int type;

    if(!self->editable){
        PyErr_SetString(PyExc_RuntimeError, "The acitivity object is not editable since it has been added to a simulation object.");
        return NULL;
    }

    if (!PyArg_ParseTupleAndKeywords(args, kwds, "iO|$OOOOO", kwlist, &type, &name, &tvalues, &fvalues, &slopeendpoints, &unit, &feature)){
        return NULL;
    }

    if(!PyUnicode_Check(name)){
        PyErr_SetString(PyExc_TypeError, "The name of the output has to be a string");
        return NULL;
    }

    if(unit && !PyUnicode_Check(unit)){
        PyErr_SetString(PyExc_TypeError, "The unit of the output has to be a string");
        return NULL;
    }

    switch(type){
        case CONSTANT:
            newOutput = Activity_ConstantOutput(fvalues);
            break;
        case PIECEWISE_CONSTANT:
            newOutput = Activity_PiecewiseConstantOutput(tvalues,fvalues);
            break;
        case PIECEWISE_LINEAR:
            newOutput = Activity_PiecewiseLinearOutput(tvalues,fvalues);
            break;
        case CUBIC_SPLINE:
            newOutput = Activity_CubicSplineOutput(tvalues,fvalues,slopeendpoints);
            break;
        case CUBIC_SPLINE_SLOPE:
            newOutput = Activity_CubicInterpolateSlopeOutput(tvalues,fvalues,slopeendpoints);
            break;
        default:
            PyErr_SetString(PyExc_ValueError, "Incorrect output type given, use predefined macros");
            return NULL;
    }

    if(!newOutput)
        return NULL;

    // check if compartment is set
    if(self->compartment != Py_None)
        name = PyUnicode_FromFormat("%S:%S",self->compartment, name);
    else
        Py_INCREF(name);

    PyList_Append(self->outputnames, name);
    self->nroutputs++;
    self->outputs = PyMem_Realloc(self->outputs, self->nroutputs * sizeof(ActivityOutput *));
    self->outputs[self->nroutputs-1] = newOutput;
    if(feature && PyObject_IsTrue(feature)){
        newOutput->isFeature = 1;
        Activity_AddFeature(self, newOutput, name, unit);
    }
    else 
        newOutput->isFeature = 0;

    Py_DECREF(name);

    Py_RETURN_NONE;
}

static ActivityOutput *
Activity_ConstantOutput(PyObject *constant){
    ConstantOutput *newOutput;
    double c;

    // check constant = fvalues
    if(!constant){
        PyErr_SetString(PyExc_ValueError, "No constant value given, please specify a constant using the 'fvalues' argument");
        return NULL;
    }
    c = PyFloat_AsDouble(constant);
    if(PyErr_Occurred()){
        PyErr_Clear();
        PyErr_SetString(PyExc_ValueError, "Constant needs to be a double");
        return NULL;
    }

    newOutput = PyMem_Malloc(sizeof(ConstantOutput));
    newOutput->base.type = CONSTANT;
    newOutput->base.out_function = (OutputFunction) Constant_function;
    newOutput->base.out_dealloc = (destructor) Constant_dealloc;
    newOutput->constant = c;

    return (ActivityOutput *) newOutput;
}

static ActivityOutput *
Activity_PiecewiseConstantOutput(PyObject *tvalues, PyObject *fvalues){
    PiecewiseOutput *newOutput;

    // check tvalues fvalues
    if(!tvalues){
        PyErr_SetString(PyExc_ValueError, "No t-values given, please specify t-values using the 'tvalues' argument");
        return NULL;
    }
    if(!fvalues){
        PyErr_SetString(PyExc_ValueError, "No f-values given, please specify f-values using the 'fvalues' argument");
        return NULL;
    }

    tvalues = PyArray_FROM_OTF(tvalues, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!tvalues){
        PyErr_SetString(PyExc_ValueError, "Cannot convert tvalues to numpy array");
        return NULL;
    }
    fvalues = PyArray_FROM_OTF(fvalues, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!fvalues){
        Py_DECREF(tvalues);
        PyErr_SetString(PyExc_ValueError, "Cannot convert fvalues to numpy array");
        return NULL;
    }

    if(PYSIZE(fvalues) != PYSIZE(tvalues) + 1){
        Py_DECREF(tvalues);
        Py_DECREF(fvalues);
        PyErr_SetString(PyExc_ValueError, "Incorrect number of t- and f-values");
        return NULL;
    }

    newOutput = PyMem_Malloc(sizeof(PiecewiseOutput));
    newOutput->base.type = PIECEWISE_CONSTANT;
    newOutput->base.out_function = (OutputFunction) PiecewiseConstant_function;
    newOutput->base.out_dealloc = (destructor) Piecewise_dealloc;
    newOutput->tvalues = tvalues;
    newOutput->fvalues = fvalues;

    return (ActivityOutput *) newOutput;
}

static ActivityOutput *
Activity_PiecewiseLinearOutput(PyObject *tvalues, PyObject *fvalues){
    PiecewiseOutput *newOutput;

    // check tvalues fvalues
    if(!tvalues){
        PyErr_SetString(PyExc_ValueError, "No t-values given, please specify t-values using the 'tvalues' argument");
        return NULL;
    }
    if(!fvalues){
        PyErr_SetString(PyExc_ValueError, "No f-values given, please specify f-values using the 'fvalues' argument");
        return NULL;
    }

    tvalues = PyArray_FROM_OTF(tvalues, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!tvalues){
        PyErr_SetString(PyExc_ValueError, "Cannot convert tvalues to numpy array");
        return NULL;
    }
    fvalues = PyArray_FROM_OTF(fvalues, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!fvalues){
        Py_DECREF(tvalues);
        PyErr_SetString(PyExc_ValueError, "Cannot convert fvalues to numpy array");
        return NULL;
    }

    if(PYSIZE(tvalues) < 2 || PYSIZE(tvalues) != PYSIZE(fvalues)){
        Py_DECREF(tvalues);
        Py_DECREF(fvalues);
        PyErr_SetString(PyExc_ValueError, "Incorrect number of t- and f-values");
        return NULL;
    }

    newOutput = PyMem_Malloc(sizeof(PiecewiseOutput));
    newOutput->base.type = PIECEWISE_LINEAR;
    newOutput->base.out_function = (OutputFunction) PiecewiseLinear_function;
    newOutput->base.out_dealloc = (destructor) Piecewise_dealloc;
    newOutput->tvalues = tvalues;
    newOutput->fvalues = fvalues;

    return (ActivityOutput *) newOutput;
}

static ActivityOutput *
Activity_CubicSplineOutput(PyObject *tvalues, PyObject *fvalues, PyObject *slopeendpoints){
    int n, r;
    CubicSplineOutput *newOutput;

    // check tvalues fvalues
    if(!tvalues){
        PyErr_SetString(PyExc_ValueError, "No t-values given, please specify t-values using the 'tvalues' argument");
        return NULL;
    }
    if(!fvalues){
        PyErr_SetString(PyExc_ValueError, "No f-values given, please specify f-values using the 'fvalues' argument");
        return NULL;
    }

    tvalues = PyArray_FROM_OTF(tvalues, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!tvalues){
        PyErr_SetString(PyExc_ValueError, "Cannot convert 'tvalues' to numpy array.");
        return NULL;
    }
    fvalues = PyArray_FROM_OTF(fvalues, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!fvalues){
        Py_DECREF(tvalues);
        PyErr_SetString(PyExc_ValueError, "Cannot convert 'fvalues' to numpy array.");
        return NULL;
    }

    if(PYSIZE(tvalues) < 2 || PYSIZE(tvalues) != PYSIZE(fvalues)){
        Py_DECREF(tvalues);
        Py_DECREF(fvalues);
        PyErr_SetString(PyExc_ValueError, "Incorrect number of t and f values.");
        return NULL;
    }

    if(slopeendpoints){
        slopeendpoints = PyArray_FROM_OTF(slopeendpoints, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
        if(!slopeendpoints){
            Py_DECREF(tvalues);
            Py_DECREF(fvalues);
            PyErr_SetString(PyExc_ValueError, "Cannot convert 'slopeendpoints' to numpy array.");
            return NULL;
        }
        if(PYSIZE(slopeendpoints) != 2){
            Py_DECREF(tvalues);
            Py_DECREF(fvalues);
            Py_DECREF(slopeendpoints);
            PyErr_SetString(PyExc_ValueError, "Incorrect number of slope endpoints, 2 points needed.");
            return NULL;
        }
    }

    n = PYSIZE(tvalues);
    newOutput = PyMem_Malloc(sizeof(CubicSplineOutput));
    newOutput->base.type = CUBIC_SPLINE;
    newOutput->base.out_function = (OutputFunction) CubicInterpolate_function;
    newOutput->base.out_dealloc = (destructor) CubicInterpolate_dealloc;
    newOutput->tvalues = tvalues;
    newOutput->fvalues = fvalues;
    if(slopeendpoints)
        newOutput->slopeendpoints = slopeendpoints;
    else{
        Py_INCREF(Py_None);
        newOutput->slopeendpoints = Py_None;
    }
    newOutput->b = PyMem_Malloc(n * sizeof(double));
    newOutput->c = PyMem_Malloc(n * sizeof(double));
    newOutput->d = PyMem_Malloc(n * sizeof(double));
    newOutput->last = 0;

    if(slopeendpoints)
        r = cubicCalcCoef(n, 1, PYDATA(slopeendpoints)[0], PYDATA(slopeendpoints)[1], PYDATA(tvalues), PYDATA(fvalues), newOutput->b, newOutput->c, newOutput->d);
    else
        r = cubicCalcCoef(n, 0, 0, 0, PYDATA(tvalues), PYDATA(fvalues), newOutput->b, newOutput->c, newOutput->d);
    
    if(r < 0)
        return NULL;
    return (ActivityOutput *) newOutput;
}

static ActivityOutput *
Activity_CubicInterpolateSlopeOutput(PyObject *tvalues, PyObject *fvalues, PyObject* slopeendpoints){
    int n, r;
    CubicSplineOutput *newOutput;

    // check tvalues fvalues
    if(!tvalues){
        PyErr_SetString(PyExc_ValueError, "No t-values given, please specify t-values using the 'tvalues' argument");
        return NULL;
    }
    if(!fvalues){
        PyErr_SetString(PyExc_ValueError, "No f-values given, please specify f-values using the 'fvalues' argument");
        return NULL;
    }

    tvalues = PyArray_FROM_OTF(tvalues, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!tvalues){
        PyErr_SetString(PyExc_ValueError, "Cannot convert 'tvalues' to numpy array.");
        return NULL;
    }
    fvalues = PyArray_FROM_OTF(fvalues, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!fvalues){
        Py_DECREF(tvalues);
        PyErr_SetString(PyExc_ValueError, "Cannot convert 'fvalues' to numpy array.");
        return NULL;
    }

    if(PYSIZE(tvalues) < 2 || PYSIZE(tvalues) != PYSIZE(fvalues)){
        Py_DECREF(tvalues);
        Py_DECREF(fvalues);
        PyErr_SetString(PyExc_ValueError, "Incorrect number of t and f values.");
        return NULL;
    }

    if(slopeendpoints){
        slopeendpoints = PyArray_FROM_OTF(slopeendpoints, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
        if(!slopeendpoints){
            Py_DECREF(tvalues);
            Py_DECREF(fvalues);
            PyErr_SetString(PyExc_ValueError, "Cannot convert 'slopeendpoints' to numpy array.");
            return NULL;
        }
        if(PYSIZE(slopeendpoints) != 2){
            Py_DECREF(tvalues);
            Py_DECREF(fvalues);
            Py_DECREF(slopeendpoints);
            PyErr_SetString(PyExc_ValueError, "Incorrect number of slope endpoints, 2 points needed.");
            return NULL;
        }
    }

    n = PYSIZE(tvalues);
    newOutput = PyMem_Malloc(sizeof(CubicSplineOutput));
    newOutput->base.type = CUBIC_SPLINE_SLOPE;
    newOutput->base.out_function = (OutputFunction) CubicInterpolateSlope_function;
    newOutput->base.out_dealloc = (destructor) CubicInterpolate_dealloc;
    newOutput->tvalues = tvalues;
    newOutput->fvalues = fvalues;
    newOutput->b = PyMem_Malloc(n * sizeof(double));
    newOutput->c = PyMem_Malloc(n * sizeof(double));
    newOutput->d = PyMem_Malloc(n * sizeof(double));
    newOutput->last = 0;

    if(slopeendpoints){
        r = cubicCalcCoef(n, 1, PYDATA(slopeendpoints)[0], PYDATA(slopeendpoints)[1], PYDATA(tvalues), PYDATA(fvalues), newOutput->b, newOutput->c, newOutput->d);
        Py_DECREF(slopeendpoints);
    } else
        r = cubicCalcCoef(n, 0, 0, 0, PYDATA(tvalues), PYDATA(fvalues), newOutput->b, newOutput->c, newOutput->d);
    if(r < 0)
        return NULL;
    return (ActivityOutput *) newOutput;
}

static PyObject *
Activity_Output(ActivityObject *self, PyObject *args, PyObject *kwds){
    int k;
    npy_intp dims[2];
    double *times, *outputdata;
    static char *kwlist[] = {"timevector", NULL};
    PyObject *timevector = NULL, *ret;

    if(!PyArg_ParseTupleAndKeywords(args, kwds, "O", kwlist, &timevector)){
        return NULL;
    }

    timevector = PyArray_FROM_OTF(timevector, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!timevector){
        PyErr_SetString(PyExc_TypeError, "Could not convert 'timevector' to numpy array");
        return NULL;
    }

    times = PYDATA(timevector);
    outputdata = (double *) malloc(self->nroutputs * PYSIZE(timevector) * sizeof(double));
    for(k = 0; k < PYSIZE(timevector); k++){
        Activity_outputFeature((PyObject *)self, times[k], &outputdata[k * self->nroutputs], NULL, -1);
    }
    
    dims[0] = PYSIZE(timevector);
    dims[1] = self->nroutputs;
    ret = PyArray_SimpleNewFromData(2, dims, NPY_DOUBLE, (void *) outputdata);
    PyArray_ENABLEFLAGS((PyArrayObject *) ret, NPY_ARRAY_OWNDATA);

    Py_DECREF(timevector);

    return ret;
}

static PyObject *
Activity_Reduce(ActivityObject *self)
{
    int outputIndex, featureIndex;
    PyObject *ret, *args, *state, *kwds, *tmp;
    ActivityOutput *AO;
    ConstantOutput *CO;
    PiecewiseOutput *PO;
    CubicSplineOutput *SO;

    state = PyList_New(self->nroutputs);
    for(outputIndex = 0, featureIndex = 0; outputIndex < self->nroutputs; outputIndex++){
        AO = self->outputs[outputIndex];

        // Create keyword arguments for re-creating output
        kwds = PyDict_New();
        tmp = PyLong_FromLong(AO->type);
        PyDict_SetItemString(kwds, "type", tmp);
        Py_DECREF(tmp);
        tmp = PyList_GetItem(self->outputnames, outputIndex);
        PyDict_SetItemString(kwds, "name", tmp);
        if(AO->isFeature){
            PyDict_SetItemString(kwds, "feature", Py_True);
            tmp = PyList_GetItem(self->featureunits, featureIndex);
            PyDict_SetItemString(kwds, "unit", tmp);
            featureIndex++;
        }
        else{
            PyDict_SetItemString(kwds, "feature", Py_False);
            PyDict_SetItemString(kwds, "unit", Py_None);
        }
        // Different arguments depending on type
        switch(AO->type){
            case CONSTANT:
                CO = (ConstantOutput *) AO;
                PyDict_SetItemString(kwds, "tvalues", Py_None);
                tmp = PyFloat_FromDouble(CO->constant);
                PyDict_SetItemString(kwds, "fvalues", tmp);
                Py_DECREF(tmp);
                PyDict_SetItemString(kwds, "slopeendpoints", Py_None);
                break;
            case PIECEWISE_CONSTANT:
            case PIECEWISE_LINEAR:
                PO = (PiecewiseOutput *) AO;
                PyDict_SetItemString(kwds, "tvalues", PO->tvalues);
                PyDict_SetItemString(kwds, "fvalues", PO->fvalues);
                PyDict_SetItemString(kwds, "slopeendpoints", Py_None);
                break;
            case CUBIC_SPLINE:
            case CUBIC_SPLINE_SLOPE:
                SO = (CubicSplineOutput *) AO;
                PyDict_SetItemString(kwds, "tvalues", SO->tvalues);
                PyDict_SetItemString(kwds, "fvalues", SO->fvalues);
                PyDict_SetItemString(kwds, "slopeendpoints", SO->slopeendpoints);
                break;
        }

        PyList_SetItem(state, outputIndex, kwds);
    }    
    
    args = Py_BuildValue("(O)", self->timeunit);
    ret = Py_BuildValue("OOO", Py_TYPE(self), args, state);
    Py_DECREF(args);
    Py_DECREF(state);
    return ret;
}

static PyObject *
Activity_SetState(ActivityObject *self, PyObject *statetuple)
{
    int k,l;
    PyObject *kwds, *state, *tmp, *args;
    char *keys[] = {"type", "name", "tvalues", "fvalues", "slopeendpoints", "unit", "feature", NULL};

    if(!PyArg_Parse(statetuple, "(O,)", &state))
        return NULL;

    if(!PyList_Check(state)){
        PyErr_SetString(PyExc_TypeError, "State needs to be a list type");
        return NULL;
    }

    // Add outputs
    args = PyTuple_New(0);
    for(k = 0; k < PyList_Size(state); k++){
        kwds = PyList_GetItem(state, k);
        if(!PyDict_Check(kwds)){
            PyErr_SetString(PyExc_TypeError, "State needs to be a list of keyword argument");
            return NULL;
        }
        // Loop through keywords and remove Py_None values
        l = 0;
        while(keys[l]){
            tmp = PyDict_GetItemString(kwds, keys[l]);
            if(!tmp){
                PyErr_SetObject(PyExc_TypeError, PyUnicode_FromFormat("Keyword '%s' missing in state list", keys[l]));
                return NULL;
            }
            if(tmp == Py_None)
                PyDict_DelItemString(kwds, keys[l]);
            l++;
        }

        Activity_AddOutput(self, args, kwds);
    }
    Py_DECREF(args);

    Py_RETURN_NONE;
}

static PyMethodDef Activity_methods[] = {
    {"AddOutput", (PyCFunction) Activity_AddOutput, METH_VARARGS | METH_KEYWORDS,
     "Add new output"},
     {"Output", (PyCFunction) Activity_Output, METH_VARARGS | METH_KEYWORDS,
     "Compute activity's outputs"},
     {"__reduce__", (PyCFunction) Activity_Reduce, METH_NOARGS,
     "__reduce__ function"},
     {"__setstate__", (PyCFunction) Activity_SetState, METH_VARARGS,
     "__setstate__ function"},
    {NULL}  /* Sentinel */
};

/*
==========================================================================================
ActivityType definition
==========================================================================================
*/

static int
Activity_init(ActivityObject *self, PyObject *args, PyObject *kwds){
    static char *kwlist[] = {"timeunit", "compartment", NULL};
    PyObject *timeunit = NULL, *compartment = NULL;
    
    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|O", kwlist, &timeunit, &compartment))
        return -1;

    if(!PyUnicode_Check(timeunit)){
        PyErr_SetString(PyExc_TypeError, "The time unit has to be a string");
        return -1;
    } else{
        Py_INCREF(timeunit);
        self->timeunit = timeunit;
    }

    if(!compartment){
        Py_INCREF(Py_None);
        self->compartment = Py_None;
    }
    else{
        if(!PyUnicode_Check(compartment)){
            PyErr_SetString(PyExc_TypeError, "The compartment has to be a string");
            return -1;
        }
        Py_INCREF(compartment);
        self->compartment = compartment;
    }

	self->nroutputs = 0;
    self->nrfeatures = 0;
    self->editable = 1;
    self->outputnames = StringList_New(0,0);
    self->featurenames = StringList_New(0,0);
    self->featureunits = StringList_New(0,0);
	self->outputs = NULL;
    self->features = NULL;
	return 0;
}

static void
Constant_dealloc(ConstantOutput *self){
    PyMem_Free(self);
}

static void
Piecewise_dealloc(PiecewiseOutput *self){
    Py_XDECREF(self->tvalues);
    Py_XDECREF(self->fvalues);
    PyMem_Free(self);
}

static void
CubicInterpolate_dealloc(CubicSplineOutput *self){
    Py_XDECREF(self->tvalues);
    Py_XDECREF(self->fvalues);
    Py_XDECREF(self->slopeendpoints);
    PyMem_Free(self->b);
    PyMem_Free(self->c);
    PyMem_Free(self->d);
    PyMem_Free(self);
}

static void
Activity_dealloc(ActivityObject *self)
{
    int k;
    Py_XDECREF(self->outputnames);
    Py_XDECREF(self->featurenames);
    Py_XDECREF(self->featureunits);
    Py_XDECREF(self->timeunit);
    Py_XDECREF(self->compartment);
    for(k = 0; k < self->nroutputs; k++){
        self->outputs[k]->out_dealloc((PyObject *) self->outputs[k]);
    }
    PyMem_Free(self->outputs);
    PyMem_Free(self->features);
    Py_TYPE(self)->tp_free((PyObject *) self);
}

static PyTypeObject ActivityType = {
    PyVarObject_HEAD_INIT(NULL, 0)
    .tp_name = "MeDigiT._Activity.Activity",
    .tp_doc = "Activity object",
    .tp_basicsize = sizeof(ActivityObject),
    .tp_itemsize = 0,
    .tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
    .tp_init = (initproc) Activity_init,
    .tp_dealloc = (destructor) Activity_dealloc,
    .tp_methods = Activity_methods,
    .tp_getset = Activity_getsetters,
    .tp_new = PyType_GenericNew,
};

/*
==========================================================================================
C_API FUNCTIONS
==========================================================================================
*/
static void
Activity_outputFeature(PyObject *tmp, double time, double *outputvector, double *featurevector, int DOflag){
    ActivityObject* self;
    int k;

    self = (ActivityObject *) tmp;
    for(k = 0; k < self->nroutputs; k++){
        self->outputs[k]->output = self->outputs[k]->out_function(self->outputs[k], time);
        outputvector[k] = self->outputs[k]->output;
    }
    // check flag
    if(DOflag == DOFLAG_FEATURE){
        for(k = 0; k < self->nrfeatures; k++){
            featurevector[k] = self->features[k]->output;
        }
    }
}

static int
Activity_nrOutputs(PyObject *tmp){
    ActivityObject* self;

    self = (ActivityObject *) tmp;
    return self->nroutputs;
}

static int
Activity_nrFeatures(PyObject *tmp){
    ActivityObject* self;

    self = (ActivityObject *) tmp;
    return self->nrfeatures;
}

static PyObject *
Activity_outputNames(PyObject *tmp){
    return ((ActivityObject *) tmp)->outputnames;
}

static PyObject *
Activity_featureNames(PyObject *tmp){
    return ((ActivityObject *) tmp)->featurenames;
}

static PyObject *
Activity_featureUnits(PyObject *tmp){
    return ((ActivityObject *) tmp)->featureunits;
}

static int 
Activity_isActivity(PyObject *obj){
    if(Py_TYPE(obj) == &ActivityType)
        return 1;
    else
        return 0;
}

static PyObject *
Activity_timeUnit(PyObject *obj){
    return ((ActivityObject *) obj)->timeunit;
}

static void
Activity_setNonEditable(PyObject *obj){
    ((ActivityObject *) obj)->editable = 0;
}

/*
==========================================================================================
Activity module definition
==========================================================================================
*/

static PyModuleDef ActivityModule = {
    PyModuleDef_HEAD_INIT,
    .m_name = "MeDigiT._Activity",
    .m_doc = "Activity Module",
    .m_size = -1
};

PyMODINIT_FUNC
PyInit__Activity(void)
{
    PyObject *m;

    static void *Activity_API[Activity_API_pointers];
    PyObject *c_api_object;

    m = PyModule_Create(&ActivityModule);
    if (m == NULL)
        return NULL;

    /* Initialize the C API pointer array */
    Activity_API[Activity_isActivity_NUM] = (void *)Activity_isActivity;
    Activity_API[Activity_outputFeature_NUM] = (void *)Activity_outputFeature;
    Activity_API[Activity_nrOutputs_NUM] = (void *)Activity_nrOutputs;
    Activity_API[Activity_nrFeatures_NUM] = (void *)Activity_nrFeatures;
    Activity_API[Activity_outputNames_NUM] = (void *)Activity_outputNames;
    Activity_API[Activity_featureNames_NUM] = (void *)Activity_featureNames;
    Activity_API[Activity_featureUnits_NUM] = (void *)Activity_featureUnits;
    Activity_API[Activity_timeUnit_NUM] = (void *)Activity_timeUnit;
    Activity_API[Activity_setNonEditable_NUM] = (void *)Activity_setNonEditable;

    /* Create a Capsule containing the API pointer array's address */
    c_api_object = PyCapsule_New((void *)Activity_API, "MeDigiT._Activity._C_API", NULL);

    // Activity/Simulation
    if (PyType_Ready(&ActivityType) < 0){
        Py_DECREF(m);
        return NULL;
    }

    Py_INCREF(&ActivityType);

    if (PyModule_AddObject(m, "Activity", (PyObject *) &ActivityType) < 0 ||
            PyModule_AddObject(m, "_C_API", c_api_object) < 0 || 
                PyModule_AddIntMacro(m,CONSTANT) < 0 ||
                    PyModule_AddIntMacro(m,PIECEWISE_CONSTANT) < 0 ||
                        PyModule_AddIntMacro(m,PIECEWISE_LINEAR) < 0 ||
                            PyModule_AddIntMacro(m,CUBIC_SPLINE) < 0 ||
                                PyModule_AddIntMacro(m,CUBIC_SPLINE_SLOPE) < 0) {
        Py_DECREF(&ActivityType);
        Py_XDECREF(c_api_object);
        Py_DECREF(m);
        return NULL;
    }

    import_array();
    import_StringList();
    return m;
}

