#define _MODELS_C
#include <Python.h>
#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
#include <numpy/arrayobject.h>
#include <math.h>

#include "_Models.h"
#include "_StringList.h"
#include "pyarraymacros.h"
//#include "../include/pymathaddon.h"
//#include "../include/kineticformulas.h"

//#include "_Models_IMPORT.h"

/*
==========================================================================================
Helper functions
==========================================================================================
*/
static int checkEquivArr(PyObject *Arr1, PyObject *Arr2){
    if(!PyArray_EquivArrTypes(PYARRAY(Arr1), PYARRAY(Arr2)) || 
            PyArray_NDIM(PYARRAY(Arr1)) != PyArray_NDIM(PYARRAY(Arr2)) || 
                PyArray_DIMS(PYARRAY(Arr1))[0] != PyArray_DIMS(PYARRAY(Arr2))[0])
        return -1;
    else
        return 0;
}

// allocation of model object
int
Model_alloc(ModelObject *self){
    npy_intp dims[1];
    // name
    Py_INCREF(Py_None);
    self->name = Py_None;
    // compartment
    Py_INCREF(Py_None);
    self->compartment = Py_None;
    // parametervalues
    dims[0] = self->model->numberof[MODEL_PARAMETER];
    self->parametervalues = PyArray_ZEROS(1, dims, NPY_DOUBLE, 0);
    if (!self->parametervalues)
        return -1;
    // statevalues
    dims[0] = self->model->numberof[MODEL_STATE];
    self->statevalues = PyArray_ZEROS(1, dims, NPY_DOUBLE, 0);
    if (!self->statevalues)
        return -1;
    // derivativevalues
    self->derivativevalues = PyArray_ZEROS(1, dims, NPY_DOUBLE, 0);
    if (!self->derivativevalues)
        return -1;
    // idvector
    self->idvector = PyArray_ZEROS(1, dims, NPY_DOUBLE, 0);
    if (!self->idvector)
        return -1;
    // statenames
    self->statenames = StringList_New(self->model->numberof[MODEL_STATE],0);
    if (!self->statenames)
        return -1;
    // featurenames
    self->featurenames = StringList_New(self->model->numberof[MODEL_FEATURE],0);
    if (!self->featurenames)
        return -1;
    // featureunits
    self->featureunits = StringList_New(self->model->numberof[MODEL_FEATURE],1);
    if (!self->featureunits)
        return -1;
    // outputnames
    self->outputnames = StringList_New(self->model->numberof[MODEL_OUTPUT],0);
    if (!self->outputnames)
        return -1;
    // inputnames
    self->inputnames = StringList_New(self->model->numberof[MODEL_INPUT],0);
    if (!self->inputnames)
        return -1;
    // parameternames
    self->parameternames = StringList_New(self->model->numberof[MODEL_PARAMETER],0);
    if (!self->parameternames)
        return -1;
    // eventnames
    self->eventnames = StringList_New(self->model->numberof[MODEL_EVENT],0);
    if (!self->eventnames)
        return -1;
    // timeunit
    self->timeunit = PyUnicode_FromString(self->model->timeunit);

    return 0;
}

/*
==========================================================================================
GETTERS AND SETTERS
==========================================================================================
*/
static PyObject *
Model_GetName(ModelObject *self, void *closure)
{
    Py_INCREF(self->name);
    return self->name;
}

static int
Model_SetName(ModelObject *self, PyObject *value, void *closure)
{
    PyObject *tmp;
    if (value == NULL) {
        PyErr_SetString(PyExc_TypeError, "Cannot delete the name attribute");
        return -1;
    }

    if (!PyUnicode_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "Only strings can be assigned the name attribute");
        return -1;
    }
    tmp = self->name;
    Py_INCREF(value);
    self->name = value;
    Py_DECREF(tmp);
    return 0;
}

static PyObject *
Model_GetCompartment(ModelObject *self, void *closure)
{
    Py_INCREF(self->compartment);
    return self->compartment;
}

static PyObject *
Model_GetTimeUnit(ModelObject *self, void *closure)
{
    Py_INCREF(self->timeunit);
    return self->timeunit;
}

static PyObject *
Model_GetStateValues(ModelObject *self, void *closure)
{
    Py_INCREF(self->statevalues);
    return self->statevalues;
}

static int
Model_SetStateValues(ModelObject *self, PyObject *value, void *closure)
{
    PyObject *tmp;
    if (value == NULL) {
        PyErr_SetString(PyExc_TypeError, "Cannot delete the statevalues attribute");
        return -1;
    }

    value = PyArray_FROM_OTF(value, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if (!value || checkEquivArr(value, self->statevalues) < 0) {
        PyErr_SetString(PyExc_TypeError,
                        "The given value does not match the statevalues attribute type");
        Py_XDECREF(value);
        return -1;
    }
    tmp = self->statevalues;
    self->statevalues = value;
    Py_DECREF(tmp);
    return 0;
}

static PyObject *
Model_GetDerivativeValues(ModelObject *self, void *closure)
{
    Py_INCREF(self->derivativevalues);
    return self->derivativevalues;
}

static int
Model_SetDerivativeValues(ModelObject *self, PyObject *value, void *closure)
{
    PyObject *tmp;
    if (value == NULL) {
        PyErr_SetString(PyExc_TypeError, "Cannot delete the derivativevalues attribute");
        return -1;
    }

    value = PyArray_FROM_OTF(value, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if (!value || checkEquivArr(value, self->derivativevalues) < 0) {
        PyErr_SetString(PyExc_TypeError,
                        "The given value does not match the derivativevalues attribute type");
        Py_XDECREF(value);
        return -1;
    }
    tmp = self->derivativevalues;
    self->derivativevalues = value;
    Py_DECREF(tmp);
    return 0;
}

static PyObject *
Model_GetParameterValues(ModelObject *self, void *closure)
{
    Py_INCREF(self->parametervalues);
    return self->parametervalues;
}

static int
Model_SetParameterValues(ModelObject *self, PyObject *value, void *closure)
{
    PyObject *tmp;
    if (value == NULL) {
        PyErr_SetString(PyExc_TypeError, "Cannot delete the parametervalues attribute");
        return -1;
    }

    value = PyArray_FROM_OTF(value, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if (!value || checkEquivArr(value, self->parametervalues) < 0) {
        PyErr_SetString(PyExc_TypeError,
                        "The given value does not match the parametervalues attribute type");
        Py_XDECREF(value);
        return -1;
    }
    tmp = self->parametervalues;
    self->parametervalues = value;
    Py_DECREF(tmp);
    return 0;
}

static PyObject *
Model_GetParameterNames(ModelObject *self, void *closure)
{
    Py_INCREF(self->parameternames);
    return self->parameternames;
}

static int
Model_SetParameterNames(ModelObject *self, PyObject *value, void *closure)
{
    return StringList_Update(self->parameternames, value);
}

static PyObject *
Model_GetStateNames(ModelObject *self, void *closure)
{
    Py_INCREF(self->statenames);
    return self->statenames;
}

static int
Model_SetStateNames(ModelObject *self, PyObject *value, void *closure)
{
    return StringList_Update(self->statenames, value);
}

static PyObject *
Model_GetFeatures(ModelObject *self, void *closure)
{
    Py_INCREF(self->featurenames);
    return self->featurenames;
}

static int
Model_SetFeaturesNames(ModelObject *self, PyObject *value, void *closure)
{
    return StringList_Update(self->featurenames, value);
}

static PyObject *
Model_GetFeatureUnits(ModelObject *self, void *closure)
{
    Py_INCREF(self->featureunits);
    return self->featureunits;
}

static PyObject *
Model_GetOutputNames(ModelObject *self, void *closure)
{
    Py_INCREF(self->outputnames);
    return self->outputnames;
}

static int
Model_SetOutputNames(ModelObject *self, PyObject *value, void *closure)
{
    return StringList_Update(self->outputnames, value);
}

static PyObject *
Model_GetInputNames(ModelObject *self, void *closure)
{
    Py_INCREF(self->inputnames);
    return self->inputnames;
}

static int
Model_SetInputNames(ModelObject *self, PyObject *value, void *closure)
{
    return StringList_Update(self->inputnames, value);
}

static PyObject *
Model_GetEventNames(ModelObject *self, void *closure)
{
    Py_INCREF(self->eventnames);
    return self->eventnames;
}

static int
Model_SetEventsNames(ModelObject *self, PyObject *value, void *closure)
{
    return StringList_Update(self->eventnames, value);
}

static PyGetSetDef Model_getsetters[] = {
    {"name", (getter) Model_GetName, (setter) Model_SetName,
     "Model name", NULL},
     {"statenames", (getter) Model_GetStateNames, (setter) Model_SetStateNames,
     "Model state names", NULL},
     {"statevalues", (getter) Model_GetStateValues, (setter) Model_SetStateValues,
     "Model state values", NULL},
     {"derivativevalues", (getter) Model_GetDerivativeValues, (setter) Model_SetDerivativeValues,
     "Model state values", NULL},
     {"parametervalues", (getter) Model_GetParameterValues, (setter) Model_SetParameterValues,
     "Model parameter values", NULL},
     {"parameternames", (getter) Model_GetParameterNames, (setter) Model_SetParameterNames,
     "Model parameter names", NULL},
     {"featurenames", (getter) Model_GetFeatures, (setter) Model_SetFeaturesNames,
     "Model feature names", NULL},
     {"outputnames", (getter) Model_GetOutputNames, (setter) Model_SetOutputNames,
     "Model output names", NULL},
     {"inputnames", (getter) Model_GetInputNames, (setter) Model_SetInputNames,
     "Model input names", NULL},
     {"eventnames", (getter) Model_GetEventNames, (setter) Model_SetEventsNames,
     "Model events names", NULL},
     {"timeunit", (getter) Model_GetTimeUnit, NULL,
     "Model time unit", NULL},
     {"compartment", (getter) Model_GetCompartment, NULL,
     "Model compartment", NULL},
     {"featureunits", (getter) Model_GetFeatureUnits, NULL,
     "Model feature units", NULL},
     {NULL}  /* Sentinel */
};

/*
==========================================================================================
PyMethods
==========================================================================================
*/
static PyObject *
Simulation_ResetStatesDerivatives(ModelObject *self, PyObject *args)
{
    self->model->initialcondition(PYDATA(self->statevalues), PYDATA(self->derivativevalues), PYDATA(self->parametervalues));
    Py_RETURN_NONE;
}

static PyObject *
Model_ResetParameters(ModelObject *self, PyObject *args)
{
    memcpy(PYDATA(self->parametervalues), self->model->defaultparameters, self->model->numberof[MODEL_PARAMETER] * sizeof(double));
    Py_RETURN_NONE;
}

static PyObject *
Model_HasAlgebraicEq(ModelObject *self, PyObject *args){
    if(self->model->has_algebraic_eq)
        Py_RETURN_TRUE;
    else
        Py_RETURN_FALSE;
}

static PyObject *
Model_DifferentialStates(ModelObject *self, PyObject *args){
    return PyArray_Cast((PyArrayObject * ) self->idvector, NPY_BOOL);
}


static PyObject *
Model_MandatoryInputs(ModelObject *self, PyObject *args){
    int k;
    npy_intp dims[1];
    char *data;
    PyObject *mandatoryinputs;

    dims[0] = self->model->numberof[MODEL_INPUT];
    mandatoryinputs = PyArray_SimpleNew(1, dims, NPY_BOOL);

    data = PyArray_BYTES((PyArrayObject *) mandatoryinputs);
    for(k = 0; k < dims[0]; k++){
        if(self->model->mandatoryinputs[k] == 0)
            data[k] = 1;
        else
            data[k] = 0;
    }

    return mandatoryinputs;
}

static PyObject *
Model_Reduce(ModelObject *self)
{
    PyObject *ret, *args, *state;

    state = Py_BuildValue("OOOOOOO",self->name, self->statenames,self->parameternames,self->featurenames,self->outputnames,self->inputnames,self->eventnames);
    args = Py_BuildValue("OOOO", self->compartment, self->statevalues, self->derivativevalues, self->parametervalues);
    
    ret = Py_BuildValue("OOO", Py_TYPE(self), args, state);
    Py_DECREF(args);
    Py_DECREF(state);
    return ret;
}

static PyObject *
Model_SetState(ModelObject *self, PyObject *statetuple)
{
    PyObject *name, *statenames, *parameternames, *featurenames, *outputnames, *inputnames, *eventnames;
    if(!PyArg_Parse(statetuple, "((OOOOOOO),)", &name, &statenames, &parameternames, &featurenames, &outputnames, &inputnames, &eventnames))
        return NULL;

    if(Model_SetName(self, name, NULL) < 0)
        return NULL;
    if(Model_SetStateNames(self, statenames, NULL) < 0)
        return NULL;
    if(Model_SetParameterNames(self, parameternames, NULL) < 0)
        return NULL;
    if(Model_SetFeaturesNames(self, featurenames, NULL) < 0)
        return NULL;
    if(Model_SetOutputNames(self, outputnames, NULL) < 0)
        return NULL;
    if(Model_SetInputNames(self, inputnames, NULL) < 0)
        return NULL;
    if(Model_SetEventsNames(self, eventnames, NULL) < 0)
        return NULL;
    Py_RETURN_NONE;
}

static PyMethodDef Model_methods[] = {
     {"ResetStatesDerivatives", (PyCFunction) Simulation_ResetStatesDerivatives, METH_NOARGS,
     "Reset state and derivative values"},
     {"ResetParameters", (PyCFunction) Model_ResetParameters, METH_NOARGS,
     "Reset parameter values"},
     {"HasAlgebraicEq", (PyCFunction) Model_HasAlgebraicEq, METH_NOARGS,
     "Check if model contains algebraic equations"},
     {"DifferentialStates", (PyCFunction) Model_DifferentialStates, METH_NOARGS,
     "Return an array of boolean, holds the value True if respective state is a differential state"},
     {"MandatoryInputs", (PyCFunction) Model_MandatoryInputs, METH_NOARGS,
     "Return an array of boolean, holds the value True if respective input is mandatory for the model"},
     {"__reduce__", (PyCFunction) Model_Reduce, METH_NOARGS,
     "__reduce__ function"},
     {"__setstate__", (PyCFunction) Model_SetState, METH_VARARGS,
     "__setstate__ function"},
    {NULL}  /* Sentinel */
};

/*
==========================================================================================
Model_Type definition
==========================================================================================
*/

static int
Model_init(ModelObject *self, PyObject *args, PyObject *kwds)
{
    // check input arguments
    static char *kwlist[] = {"compartment", "statevalues", "derivativevalues", "parametervalues", NULL};
    PyObject *compartment = NULL, *statevalues = NULL, *derivativevalues = NULL, *parametervalues = NULL;
    PyObject *tmp;
    const char *name;
    int k;
    double *idvector;

    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|OOOO", kwlist, &compartment, &statevalues, &derivativevalues, &parametervalues))
        return -1;

    // parametervalues
    if (parametervalues) {
        if(Model_SetParameterValues(self, parametervalues, NULL) < 0)
            return -1;
    } else
        memcpy(PYDATA(self->parametervalues), self->model->defaultparameters, self->model->numberof[MODEL_PARAMETER] * sizeof(double));
    // statevalues and derivativevalues
    // Reset states derivative, need to be done first since both 
    // states/derivatives are effected
    self->model->initialcondition(PYDATA(self->statevalues), PYDATA(self->derivativevalues), PYDATA(self->parametervalues));
    if (statevalues) {
        if(Model_SetStateValues(self, statevalues, NULL) < 0)
            return -1;
    } 
    if (derivativevalues) {
        if(Model_SetDerivativeValues(self, derivativevalues, NULL) < 0)
            return -1;
    }
    // idvector
    // 1.0: state is differential, 0.0: state is algebraic
    idvector = PYDATA(self->idvector);
    for(k = 1; k < self->model->differentialstates[0] + 1; k++){
        idvector[self->model->differentialstates[k]] = 1.0;
    }
    // compartment
    if (compartment && compartment != Py_None) {
        if(!PyUnicode_Check(compartment)){
            PyErr_SetString(PyExc_TypeError, "Compartment needs to be a string object");
            return -1;
        }
        tmp = self->compartment;
        Py_INCREF(compartment);
        self->compartment = compartment;
        Py_DECREF(tmp);
    }
    // name
    name = Py_TYPE(self)->tp_name; // = MeDigiT.Models.modelname
    name = strrchr(name,'.') + 1; // = modelname
    tmp = self->name;
    self->name = self->compartment != Py_None ? PyUnicode_FromFormat("%S:%s",self->compartment, name) : PyUnicode_FromString(name);
    Py_DECREF(tmp);
    // statenames
    for(k = 0; k < self->model->numberof[MODEL_STATE]; k++){
        tmp = self->compartment != Py_None ? PyUnicode_FromFormat("%S:%s",self->compartment, self->model->statenames[k]) : PyUnicode_FromString(self->model->statenames[k]);
        if(!tmp || PyList_SetItem(self->statenames, k, tmp) < 0){
            Py_XDECREF(tmp);
            return -1;
        }
    }
    // featurenames
    for(k = 0; k < self->model->numberof[MODEL_FEATURE]; k++){
        tmp = self->compartment != Py_None ? PyUnicode_FromFormat("%S:%s",self->compartment, self->model->featurenames[k]) : PyUnicode_FromString(self->model->featurenames[k]);
        if(!tmp || PyList_SetItem(self->featurenames, k, tmp) < 0){
            Py_XDECREF(tmp);
            return -1;
        }
    }
    // featureunits
    for(k = 0; k < self->model->numberof[MODEL_FEATURE]; k++){
        tmp = self->compartment != Py_None ? PyUnicode_FromFormat("%S:%s",self->compartment, self->model->featureunits[k]) : PyUnicode_FromString(self->model->featureunits[k]);
        if(!tmp || PyList_SetItem(self->featureunits, k, tmp) < 0){
            Py_XDECREF(tmp);
            return -1;
        }
    }
    // outputnames
    for(k = 0; k < self->model->numberof[MODEL_OUTPUT]; k++){
        tmp = self->compartment != Py_None ? PyUnicode_FromFormat("%S:%s",self->compartment, self->model->outputnames[k]) : PyUnicode_FromString(self->model->outputnames[k]);
        if(!tmp || PyList_SetItem(self->outputnames, k, tmp) < 0){
            Py_XDECREF(tmp);
            return -1;
        }
    }
    // inputnames
    for(k = 0; k < self->model->numberof[MODEL_INPUT]; k++){
        tmp = self->compartment != Py_None ? PyUnicode_FromFormat("%S:%s",self->compartment, self->model->inputnames[k]) : PyUnicode_FromString(self->model->inputnames[k]);
        if(!tmp || PyList_SetItem(self->inputnames, k, tmp) < 0){
            Py_XDECREF(tmp);
            return -1;
        }
    }
    // parameternames
    for(k = 0; k < self->model->numberof[MODEL_PARAMETER]; k++){
        tmp = self->compartment != Py_None ? PyUnicode_FromFormat("%S:%s",self->compartment, self->model->parameternames[k]) : PyUnicode_FromString(self->model->parameternames[k]);
        if(!tmp || PyList_SetItem(self->parameternames, k, tmp) < 0){
            Py_XDECREF(tmp);
            return -1;
        }
    }
    // eventnames
    for(k = 0; k < self->model->numberof[MODEL_EVENT]; k++){
        tmp = self->compartment != Py_None ? PyUnicode_FromFormat("%S:%s",self->compartment, self->model->eventnames[k]) : PyUnicode_FromString(self->model->eventnames[k]);
        if(!tmp || PyList_SetItem(self->eventnames, k, tmp) < 0){
            Py_XDECREF(tmp);
            return -1;
        }
    }

    return 0;
}

static void
Model_dealloc(ModelObject *self)
{
    Py_XDECREF(self->name);
    Py_XDECREF(self->compartment);
    Py_XDECREF(self->statevalues);
    Py_XDECREF(self->derivativevalues);
    Py_XDECREF(self->idvector);
    Py_XDECREF(self->parametervalues);
    Py_XDECREF(self->statenames);
    Py_XDECREF(self->featurenames);
    Py_XDECREF(self->featureunits);
    Py_XDECREF(self->outputnames);
    Py_XDECREF(self->inputnames);
    Py_XDECREF(self->parameternames);
    Py_XDECREF(self->eventnames);
    Py_XDECREF(self->timeunit);
    Py_TYPE(self)->tp_free((PyObject *) self);
}

static PyTypeObject Model_Type = {
    PyVarObject_HEAD_INIT(NULL, 0)
    .tp_name = "MeDigiT._Models.Model",
    .tp_doc = "Model object",
    .tp_basicsize = sizeof(ModelObject),
    .tp_itemsize = 0,
    .tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
    .tp_init = (initproc) Model_init,
    .tp_dealloc = (destructor) Model_dealloc,
    .tp_getset = Model_getsetters,
    .tp_methods = Model_methods,
};

/*
==========================================================================================
C_API
==========================================================================================
*/
static int 
Model_isModel(PyObject *obj){
    if(Py_TYPE(obj)->tp_base == &Model_Type)
        return 1;
    else
        return 0;
}

static ModelFunction * 
Model_modelFunction(PyObject *obj){
    return ((ModelObject *) obj)->model->function;
}

static PyObject * 
Model_parameters(PyObject *obj){
    return ((ModelObject *) obj)->parametervalues;
}

static PyObject *
Model_timeUnit(PyObject *obj){
    return ((ModelObject *) obj)->timeunit;
}

static PyObject * 
Model_name(PyObject *obj){
    return ((ModelObject *) obj)->name;
}

static const int * 
Model_numberof(PyObject *obj){
    return ((ModelObject *) obj)->model->numberof;
}

static const int * 
Model_inputDependency(PyObject *obj){
    return ((ModelObject *) obj)->model->inputdependency;
}

static PyObject * 
Model_stateNames(PyObject *obj){
    return ((ModelObject *) obj)->statenames;
}

static PyObject * 
Model_stateValues(PyObject *obj){
    return ((ModelObject *) obj)->statevalues;
}

static PyObject * 
Model_derivativeValues(PyObject *obj){
    return ((ModelObject *) obj)->derivativevalues;
}

static PyObject * 
Model_featureNames(PyObject *obj){
    return ((ModelObject *) obj)->featurenames;
}

static PyObject * 
Model_featureUnits(PyObject *obj){
    return ((ModelObject *) obj)->featureunits;
}

static PyObject * 
Model_outputNames(PyObject *obj){
    return ((ModelObject *) obj)->outputnames;
}

static PyObject * 
Model_inputNames(PyObject *obj){
    return ((ModelObject *) obj)->inputnames;
}

static PyObject * 
Model_parameterNames(PyObject *obj){
    return ((ModelObject *) obj)->parameternames;
}

static PyObject * 
Model_eventNames(PyObject *obj){
    return ((ModelObject *) obj)->eventnames;
}

static PyObject * 
Model_idVector(PyObject *obj){
    return ((ModelObject *) obj)->idvector;
}

static const int
Model_hasAlgebraicEq(PyObject *obj){
    return ((ModelObject *) obj)->model->has_algebraic_eq;
}

static const int * 
Model_mandatoryInputs(PyObject *obj){
    return ((ModelObject *) obj)->model->mandatoryinputs;
}

/*
==========================================================================================
Model module definitions
==========================================================================================
*/
static PyModuleDef ModelsModule = {
    PyModuleDef_HEAD_INIT,
    .m_name = "MeDigiT._Models",
    .m_doc = "Model definition module",
    .m_size = -1
};

PyMODINIT_FUNC
PyInit__Models(void)
{
    PyObject *m;

    static void *Models_C_API[Models_C_API_pointers];
    static void *Models_M_API[Models_M_API_pointers];
    PyObject *c_api_object, *m_api_object;

    m = PyModule_Create(&ModelsModule);
    if (m == NULL)
        return NULL;

    /* Initialize the C API pointer array */
    Models_C_API[Model_isModel_NUM] = (void *)Model_isModel;
    Models_C_API[Model_modelFunction_NUM] = (void *)Model_modelFunction;
    Models_C_API[Model_parameters_NUM] = (void *)Model_parameters;
    Models_C_API[Model_timeUnit_NUM] = (void *)Model_timeUnit;
    Models_C_API[Model_name_NUM] = (void *)Model_name;
    Models_C_API[Model_numberof_NUM] = (void *)Model_numberof;
    Models_C_API[Model_inputDependency_NUM] = (void *)Model_inputDependency;
    Models_C_API[Model_stateValues_NUM] = (void *)Model_stateValues;
    Models_C_API[Model_featureNames_NUM] = (void *)Model_featureNames;
    Models_C_API[Model_featureUnits_NUM] = (void *)Model_featureUnits;
    Models_C_API[Model_outputNames_NUM] = (void *)Model_outputNames;
    Models_C_API[Model_inputNames_NUM] = (void *)Model_inputNames;
    Models_C_API[Model_parameterNames_NUM] = (void *)Model_parameterNames;
    Models_C_API[Model_eventNames_NUM] = (void *)Model_eventNames;
    Models_C_API[Model_derivativeValues_NUM] = (void *)Model_derivativeValues;
    Models_C_API[Model_stateNames_NUM] = (void *)Model_stateNames;
    Models_C_API[Model_idVector_NUM] = (void *)Model_idVector;
    Models_C_API[Model_hasAlgebraicEq_NUM] = (void *)Model_hasAlgebraicEq;
    Models_C_API[Model_mandatoryInputs_NUM] = (void *)Model_mandatoryInputs;

    /* Initialize the M API pointer array */
    Models_M_API[Model_alloc_NUM] = (void *)Model_alloc;
    Models_M_API[Model_Base_Type_NUM] = (void *)&Model_Type;

    /* Create a Capsule containing the API pointer array's address */
    c_api_object = PyCapsule_New((void *)Models_C_API, "MeDigiT._Models._C_API", NULL);
    m_api_object = PyCapsule_New((void *)Models_M_API, "MeDigiT._Models._M_API", NULL);

    if (PyModule_AddObject(m, "_C_API", c_api_object) < 0){
        Py_XDECREF(c_api_object);
        Py_DECREF(m);
        return NULL;
    }
    if (PyModule_AddObject(m, "_M_API", m_api_object) < 0){
        Py_XDECREF(m_api_object);
        Py_XDECREF(c_api_object);
        Py_DECREF(m);
        return NULL;
    }
    //if(addModels(m, nrModels_Type, Models_Type) < 0){
    //    Py_DECREF(c_api_object);
    //    Py_DECREF(m);
    //    return NULL;
    //}
    import_array();
    import_StringList();
    return m;
}
