from ._StringList import StringList
from ._Activity import Activity,CONSTANT,PIECEWISE_CONSTANT,PIECEWISE_LINEAR,CUBIC_SPLINE,CUBIC_SPLINE_SLOPE
from . import _Models
from ._Simulation import Simulation,TimeScales
from . import tools
import os, pkgutil
import importlib
import glob
import sys
from distutils.sysconfig import get_config_var

##################################################################################################
# List all installed models
##################################################################################################
def installedModels():
	pkgpath = os.path.join(os.path.dirname(os.path.realpath(__file__)),"Models")
	return [name for _, name, _ in pkgutil.iter_modules([pkgpath])]

def installModel(file, force = False, keepcfile = False):
	if type(file) is list:
		for f in file:
			installModel(f)
	else:
		ModelsFolder = os.path.join(os.path.dirname(os.path.realpath(__file__)),"Models")
		files = glob.glob(os.path.abspath(file))
		if len(files) == 0:
			raise ValueError('Could not find file "{}"'.format(file))
		for f in files:
			try:
				model = tools._modelStructure(f)
				# if not force install, check if model is installed and modification time is less for text file than for module file
				modulefile = os.path.join(ModelsFolder,model["NAME"] + get_config_var("EXT_SUFFIX"))
				if not force and model["NAME"] in installedModels() and os.path.getmtime(f) < os.path.getmtime(modulefile):
					print(f"Model '{model['NAME']}' is already installed and up to date.")
					continue

				# Check if imported
				if ("MeDigiT.Models." + model["NAME"]) in sys.modules:
					raise ValueError(f"Model '{model['NAME']}' has already been imported into memory, you will need to restart Python before re-installing the model.")
				# Check if modulefile exist on file system and remove
				if os.path.isfile(modulefile):
					try:
						os.remove(modulefile)
					except:
						raise ValueError(f"Model '{model['NAME']}' cannot be re-installed, probably because it is being used by another process. Try terminating any process that might use the model and try again.")
				
				# create C-file and module file
				cfile = tools._modelCFile(model)
				tools._modelModuleFile(cfile)
				if not keepcfile:
					os.remove(cfile)
				print(f"Model '{model['NAME']}' succesfully installed.")
			except Exception as e:
				raise ValueError(f"Error while installing: {f}") from e

def importModel(modelname):
	try:
		module = importlib.import_module("MeDigiT.Models." + modelname)
		return getattr(module, modelname)
	except ModuleNotFoundError:
		raise ValueError(f"Error while importing model: '{modelname}'") from e

def uninstallModel(modelname=None, uninstallall=False):
	ModelsFolder = os.path.join(os.path.dirname(os.path.realpath(__file__)),"Models")
	if uninstallall:
		modulefiles = glob.glob(os.path.join(ModelsFolder,'*' + get_config_var("EXT_SUFFIX")))
		uninstallModel([os.path.basename(m).split(os.extsep)[0] for m in modulefiles])
		print("\nAll models have been uninstalled.")
	elif type(modelname) is list:
		for m in modelname:
			uninstallModel(m)
	else:
		try:
			if ("MeDigiT.Models." + modelname) in sys.modules:
				raise ValueError(f"Model '{modelname}' has been imported into memory, you will need to restart Python before uninstalling the model.")
			file = os.path.join(ModelsFolder,modelname + get_config_var("EXT_SUFFIX"))
			if not os.path.isfile(file):
				raise ValueError(f"Model file for model '{modelname}' not found")
			os.remove(file)
			print(f"Model '{modelname}' has been uninstalled.")
		except Exception as e:
			raise ValueError(f"Model '{modelname}' could not be uninstalled, probably because it is being used by another process. Try terminating any process that might use the model and try again.")



