#include <Python.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

/* Sundials includes */
#include <cvode/cvode.h>                          /* prototypes for CVODE fcts., consts.          */
#include <nvector/nvector_serial.h>               /* access to serial N_Vector                    */
#include <sunmatrix/sunmatrix_dense.h>            /* access to dense SUNMatrix                    */
#include <sunlinsol/sunlinsol_dense.h>            /* access to dense SUNLinearSolver              */
#include <sunnonlinsol/sunnonlinsol_newton.h>     /* access to the newton SUNNonlinearSolver      */
#include <sunnonlinsol/sunnonlinsol_fixedpoint.h> /* access to the fixed point SUNNonlinearSolver */
#include <sundials/sundials_types.h>              /* definition of realtype                       */
#include "SUNDIALSflags55.h"
#include "SUNDIALSinterface55.h"

/* Definitions for CVODES */
#define ABSTOL  1.0e-6  
#define RELTOL  1.0e-6  
#define MAXNUMSTEPS     100000  /* default in Sundials is 500 but higher is better for general purpose */
#define MAXERRTESTFAILS 50      /* default is 7 */

typedef struct{
    SUNDIALSObject sundialsobject;
    // Integrator setup
    void *integrator_mem;
    SUNMatrix A;
    SUNLinearSolver LS;
    SUNNonlinearSolver NLS;
    N_Vector u;
    int *eventstatus;
    // User setup
    void *user_data;
    int nrstates;
    int nrevents;
    int nrfeatures;
    // Simulations options
    int showintegratorstats;
    int minstep;
    int maxstep;
    long int maxnumsteps;
    double reltol;
    double abstol;
    int maxerrtestfails;
    int maxorder;
    int maxconvfails;
    double initstep;
    int maxnonliniters;
    int method;
}CVODEObject;

/* Internal functions */
static int CVODE55_setOptions(void *self, PyObject *options);
static int CVODE55_integrate(void *self, int nrtimesteps, double *timevector, double *statevector, double *derivativevector);
static void CVODE55_free(void *self);
static void handleCVODEOptions(CVODEObject*, PyObject*);
static void createCVODEIntegratorMemory(CVODEObject *);
static void setCVODEIntegratorOptions(CVODEObject *);
static void reportCVODEStats(void *cvode_mem);
static void freeCVODEIntegratorMemory(CVODEObject *);

/* Functions containing the model equations */
extern void model(void *simData, double time_local, double *statevector, double *derivativevector, double *RESvector, double *featurevector, int DOflag, int timeindex, double *eventvector, int *eventstatus);

/*
========================================================================
 CVODE init functions
========================================================================
*/
SUNDIALSObject * 
CVODE55_init(void *user_data, int nrstates, int nrevents, int nrfeatures, PyObject *options)
{
    CVODEObject *cvode;
    // Create SUNDIALS memory
    cvode = (CVODEObject *) malloc(sizeof(CVODEObject));

    // Simulation options
    handleCVODEOptions(cvode, options);
    if(PyErr_Occurred()){
        free(cvode);
        return NULL;
    }
    // SUNDIALS Object
    cvode->sundialsobject.integrate = CVODE55_integrate;
    cvode->sundialsobject.setoptions = CVODE55_setOptions;
    cvode->sundialsobject.free = CVODE55_free;
    cvode->sundialsobject.simdata.nreventstriggered = 0;
    cvode->sundialsobject.simdata.eventstatusdata = NULL;
    cvode->sundialsobject.simdata.eventtimedata = NULL;
    // User setup
    cvode->user_data = user_data;
    cvode->nrstates = nrstates;
    cvode->nrevents = nrevents;
    cvode->nrfeatures = nrfeatures;
    // Integrator setup
    cvode->u = N_VMake_Serial(nrstates,NULL);
    cvode->eventstatus = (int *) calloc(nrevents, sizeof(int));

    createCVODEIntegratorMemory(cvode);
    setCVODEIntegratorOptions(cvode);

    return (SUNDIALSObject *) cvode;
}

/*
========================================================================
 CVODE set options
========================================================================
*/
static int CVODE55_setOptions(void *self, PyObject *options)
{
    int old_method;
    CVODEObject *cvode;

    cvode = (CVODEObject *) self;
    old_method = cvode->method;

    handleCVODEOptions(cvode, options);
    if(PyErr_Occurred())
        return -1;
    // check if method is change, CVODE integrator memory needs to be re-created
    if(old_method != cvode->method){
        freeCVODEIntegratorMemory(cvode);
        createCVODEIntegratorMemory(cvode);
    }
    setCVODEIntegratorOptions(cvode);
    return 0;
}

/*
========================================================================
 CVODE integrate
========================================================================
*/
static int CVODE55_integrate(void *self, int nrtimesteps, double *timevector, double *statevector, double *derivativevector)
{
    int k, flag, timeindex;
    double treturn, tendstep, *eventtimedata, *featuredata;
    int nrevents, nreventstriggered, *eventstatus, nrfeatures;
    char *eventstatusdata;
    void *cvode_mem, *user_data;
    N_Vector u;
    CVODEObject *cvode;

    if(nrtimesteps <= 1){
        PyErr_SetString(PyExc_ValueError, "'timevector' input argument needs to be a vector if you want to do a simulation.");
        return -1;
    }

    cvode = (CVODEObject *) self;
    cvode_mem = cvode->integrator_mem;
    user_data = cvode->user_data;
    eventstatus = cvode->eventstatus;
    nrevents = cvode->nrevents;
    nrfeatures = cvode->nrfeatures;
    u = cvode->u;
    N_VSetArrayPointer(statevector, u);

    /* Simulations data */
    nreventstriggered = 0;
    eventstatusdata = NULL;
    eventtimedata = NULL;
    featuredata = (double *) malloc(nrtimesteps * nrfeatures * sizeof(double));

    CVodeReInit(cvode_mem, timevector[0], u);
    // calculate feaures fore time point 0
    model(user_data, timevector[0], statevector, NULL, NULL, featuredata, DOFLAG_FEATURE, 0, NULL, NULL);
    // Simulate
    timeindex = 1;
    tendstep = timevector[1];
    while(1) {
        flag = CVode(cvode_mem, tendstep, u, &treturn, CV_NORMAL);
        if (flag < 0) {
            if (flag == CV_TOO_MUCH_WORK)
                PyErr_SetString(PyExc_RuntimeError, "CVODE Error: CV_TOO_MUCH_WORK");
            else if (flag == CV_TOO_MUCH_ACC)
                PyErr_SetString(PyExc_RuntimeError, "CVODE Error: CV_TOO_MUCH_ACC");
            else if (flag == CV_ERR_FAILURE || flag == CV_CONV_FAILURE)
                PyErr_SetString(PyExc_RuntimeError, "CVODE Error: CV_ERR_FAILURE");
            else
                PyErr_SetObject(PyExc_RuntimeError, PyUnicode_FromFormat("CVODE Error Flag: %d", flag));
            free(eventstatusdata);
            free(eventtimedata);
            free(featuredata);
            return -1;
        }
        if (flag == CV_ROOT_RETURN) {
            /* Event happened */
            CVodeGetRootInfo(cvode_mem, eventstatus);
            model(user_data, treturn, statevector, NULL, NULL, NULL, DOFLAG_EVENTASSIGN, -1, NULL, eventstatus);
            flag = CVodeReInit(cvode_mem, treturn, u);
            nreventstriggered += 1;  
            if (nreventstriggered == 1) {
                eventtimedata = (double *) calloc(1, sizeof(double));
                eventstatusdata = (char *) calloc(nrevents, sizeof(char));
            } else {
                eventtimedata = (double *) realloc((void *) eventtimedata, nreventstriggered * sizeof(double));
                eventstatusdata = (char *) realloc((void *) eventstatusdata, nrevents * nreventstriggered * sizeof(char));
            }
            eventtimedata[nreventstriggered - 1] = treturn;
            for (k = 0; k < nrevents; k ++)
                eventstatusdata[nrevents * (nreventstriggered - 1) + k] = (char) eventstatus[k];
        } 
        if (tendstep == treturn) { 
            model(user_data, tendstep, statevector, NULL, NULL, &featuredata[timeindex * nrfeatures],DOFLAG_FEATURE, timeindex, NULL, NULL);
            
            timeindex += 1;
            if (timeindex < nrtimesteps) 
                tendstep = timevector[timeindex];
            else
                break;
        }
    }
    // Update derivativevector to last timepoint
    model(user_data, tendstep, statevector, derivativevector, NULL, NULL, DOFLAG_DDT, -1, NULL, NULL);

    // Update event data
    cvode->sundialsobject.simdata.nreventstriggered = nreventstriggered;
    cvode->sundialsobject.simdata.eventtimedata = eventtimedata;
    cvode->sundialsobject.simdata.eventstatusdata = eventstatusdata;
    cvode->sundialsobject.simdata.featuredata = featuredata;

    if(cvode->showintegratorstats){
        reportCVODEStats(cvode->integrator_mem);
    }

    return 0;
}

/*
========================================================================
 CVODE free
========================================================================
*/
static void CVODE55_free(void *self)
{
    CVODEObject *cvode = (CVODEObject *) self;
    freeCVODEIntegratorMemory(cvode);
    N_VDestroy_Serial(cvode->u);
    free(cvode->eventstatus);
    free(cvode);
}

/*
========================================================================
 RHS function f(t,u)
========================================================================
*/
static int f(double time, N_Vector u, N_Vector udot, void *user_data)
{
    double *statevector, *derivativevector;
    /* connect input and result data */
    statevector = N_VGetArrayPointer(u);
    derivativevector = N_VGetArrayPointer(udot);
    /* run the model */
    model(user_data, time, statevector, derivativevector, NULL, NULL, DOFLAG_DDT, -1, NULL, NULL);
    return(0);
}

/*
========================================================================
 Event function 
========================================================================
*/
static int g(double time, N_Vector y, double *gout, void *user_data)
{
    double *statevec;
    /* connect input data */
    statevec = N_VGetArrayPointer(y);
    /* run the event function */
    model(user_data, time, statevec, NULL, NULL, NULL, DOFLAG_EVENT, -1, gout, NULL);
    return(0);
}

/*
========================================================================
 Free integrator memory 
========================================================================
*/
static void freeCVODEIntegratorMemory(CVODEObject *cvode)
{
    /* Free all CVODE related memory */
    CVodeFree(&cvode->integrator_mem); /* Free the integrator memory */
    SUNNonlinSolFree(cvode->NLS); /* Free the non-linear solver */
    SUNLinSolFree(cvode->LS); /* Free the linear solver */
    SUNMatDestroy(cvode->A); /* Free the matrix used in linear solver */
}

/*
==============================================
INITIALIZE INTEGRATOR
==============================================
*/
static void createCVODEIntegratorMemory(CVODEObject *cvode)
{
    int k, *rootdir;
    void *integrator_mem;
    double *vector;

    /* Set integration method (stiff or non-stiff) */
    if (cvode->method == 0) {
        integrator_mem = CVodeCreate(CV_BDF);    /* default (stiff) */
        cvode->NLS = NULL; // Use default newton
        cvode->A = SUNDenseMatrix(cvode->nrstates, cvode->nrstates);
        cvode->LS = SUNLinSol_Dense(cvode->u, cvode->A);
    }
    else{
        integrator_mem = CVodeCreate(CV_ADAMS);          /* nonstiff */
        cvode->NLS = SUNNonlinSol_FixedPoint(cvode->u, 0);
        cvode->A = NULL;
        cvode->LS = NULL;
    }
    // Init
    vector = (double *) malloc(cvode->nrstates * sizeof(double)); // Dummy variable so initialization can be done
    N_VSetArrayPointer(vector, cvode->u);
    CVodeInit(integrator_mem, f, 0, cvode->u);

    // Attach solvers
    if (cvode->method == 0){
        CVodeSetLinearSolver(integrator_mem, cvode->LS, cvode->A);
        CVodeSetJacFn(integrator_mem, NULL); // Use a difference quotient Jacobian
    } else{
        CVodeSetNonlinearSolver(integrator_mem, cvode->NLS);
    }

    // Add event function
    if (cvode->nrevents > 0) {
        CVodeRootInit(integrator_mem, cvode->nrevents, g);
        // Set root dir = 1 => increasing direction
        rootdir = (int *) malloc(cvode->nrevents * sizeof(int));
        for(k = 0; k < cvode->nrevents; k++)
            rootdir[k] = 1;
        CVodeSetRootDirection(integrator_mem, rootdir);
        free(rootdir);
    }

    cvode->integrator_mem = integrator_mem;
    free(vector);
}

static void setCVODEIntegratorOptions(CVODEObject *cvode)
{
    void *integrator_mem;
    integrator_mem = cvode->integrator_mem;
    // Set options
    CVodeSStolerances(integrator_mem, cvode->reltol, cvode->abstol);
    CVodeSetUserData(integrator_mem, cvode->user_data);
    if (cvode->minstep > 0) CVodeSetMinStep(integrator_mem, cvode->minstep);
    if (cvode->maxstep > 0) CVodeSetMaxStep(integrator_mem, cvode->maxstep);
    if (cvode->maxnumsteps > 0) CVodeSetMaxNumSteps(integrator_mem, cvode->maxnumsteps);
    CVodeSetMaxErrTestFails(integrator_mem, cvode->maxerrtestfails);
    CVodeSetMaxOrd(integrator_mem, cvode->maxorder);
    CVodeSetMaxConvFails(integrator_mem, cvode->maxconvfails);
    CVodeSetInitStep(integrator_mem, cvode->initstep);
    CVodeSetMaxNonlinIters(integrator_mem, cvode->maxnonliniters);
}

/*
==============================================
 HANDLE THE SIMULATION OPTIONS
==============================================
*/
static void handleCVODEOptions(CVODEObject *cvode, PyObject *options)
{
    PyObject *item;
    /* default options */
    cvode->showintegratorstats    = 0;
    cvode->minstep                = 0.0;              /* default value in CVODE */
    cvode->maxstep                = 0.0;              /* default value in CVODE => infinity */
    cvode->maxnumsteps            = MAXNUMSTEPS;      /* we set a value different from CVODE default values as default */
    cvode->maxerrtestfails        = MAXERRTESTFAILS;  /* we set a value different from CVODE default values as default */
    cvode->maxorder               = 5;                /* default value for default method (stiff) */
    cvode->maxconvfails           = 10;               /* default value in CVODE */
    cvode->initstep               = 0.0;              /* default value in CVODE */
    cvode->maxnonliniters         = 3;                /* default value in CVODE */
    cvode->method                 = 0;                /* 0 = stiff, 1 = nonstiff, default: 0 */
    cvode->reltol                 = RELTOL;           /* we set a value different from CVODE default values as default */
    cvode->abstol                 = ABSTOL;           /* we set a value different from CVODE default values as default */
    if(PyDict_Check(options)){
        if (PyDict_Size(options) != 0) {
            item = PyDict_GetItemString(options, "showintegratorstats");
            if (item != NULL) {
                if (PyObject_IsTrue(item))
                    cvode->showintegratorstats = 1; 
                else 
                    cvode->showintegratorstats = 0;
            }
            item = PyDict_GetItemString(options, "minstep");
            if (item != NULL) {
                if (PyFloat_Check(item)) {
                    cvode->minstep = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.minstep' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "maxstep");
            if (item != NULL) {
                if (PyFloat_Check(item)) {
                    cvode->maxstep = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxstep' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "maxnumsteps");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    cvode->maxnumsteps = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxnumsteps' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "method");
            if (item != NULL) {
                if (PyUnicode_Check(item)) {
                    if (PyUnicode_GET_LENGTH(item) == 8) {
                        cvode->method      = 1;    /* "nonstiff" method ... just check the length of the string */
                        cvode->maxorder    = 12;   /* nonstiff Adams */
                    } else if (PyUnicode_GET_LENGTH(item) == 5) {
                        cvode->method      = 0;    /* "stiff method */
                        cvode->maxorder    = 5;    /* stiff BDF */
                    } else {
                        PyErr_SetString(PyExc_ValueError, "'options.method' wrongly defined.");
                    }
                } else {
                    PyErr_SetString(PyExc_ValueError, "'options.method' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "reltol");
            if (item != NULL) {
                if (PyFloat_Check(item)) {
                    cvode->reltol = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.reltol' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "abstol");
            if (item != NULL) {
                if (PyFloat_Check(item)) {
                    cvode->abstol = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.abstol' wrongly defined.");
                }
            }
            /* Additional Options */
            item = PyDict_GetItemString(options, "maxerrtestfails");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    cvode->maxerrtestfails = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxerrtestfails' wrongly defined.");         
                }
            }   
            item = PyDict_GetItemString(options, "maxorder");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    cvode->maxorder = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxorder' wrongly defined.");    
                }
            }        
            item = PyDict_GetItemString(options, "maxconvfails");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    cvode->maxconvfails = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxconvfails' wrongly defined."); 
                }
            }
            item = PyDict_GetItemString(options, "initstep");
            if (item != NULL) {
                if (PyFloat_Check(item)) {
                    cvode->initstep = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.initstep' wrongly defined.");                
                }
            }
            item = PyDict_GetItemString(options, "maxnonliniters");
            if (item != NULL) {
                if (PyFloat_Check(item)) {
                    cvode->maxnonliniters = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxnonliniters' wrongly defined.");  
                }
            }              
        }
    } else
        PyErr_SetString(PyExc_ValueError, "Options needs to be a dictionary.");
    return;
}

/*
 ========================================================================
  Integration statistics report function
 ========================================================================
 */
static void reportCVODEStats(void *integrator_mem)
{
    long int STATS_nsteps;
    long int STATS_nfevals;
    long int STATS_netfails;
    double STATS_hinused;
    double STATS_tolsfac;
    /* Get the statistics */
    CVodeGetNumSteps(integrator_mem,&STATS_nsteps);
    CVodeGetNumRhsEvals(integrator_mem,&STATS_nfevals);
    CVodeGetNumErrTestFails(integrator_mem,&STATS_netfails);
    CVodeGetActualInitStep(integrator_mem,&STATS_hinused);
    CVodeGetTolScaleFactor(integrator_mem,&STATS_tolsfac);
    /* Report statistics */
    printf("\nIntegrator Statistics\n");
    printf("=====================\n");
    printf("Cumulative number of internal steps:    %ld\n",STATS_nsteps);
    printf("No. of calls to r.h.s. function:        %ld\n",STATS_nfevals);
    printf("No. of local error test failures:       %ld\n",STATS_netfails);
    printf("Actual init step size used:             %g\n",STATS_hinused);
    printf("Suggested factor for tolerance scaling: %g\n\n",STATS_tolsfac);
}
