#include <Python.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

/* Sundials includes */
#include <ida/ida.h>                          /* prototypes for IDA fcts., consts.          */
#include <nvector/nvector_serial.h>               /* access to serial N_Vector                    */
#include <sunmatrix/sunmatrix_dense.h>            /* access to dense SUNMatrix                    */
#include <sunlinsol/sunlinsol_dense.h>            /* access to dense SUNLinearSolver              */
#include <sunnonlinsol/sunnonlinsol_newton.h>     /* access to the newton SUNNonlinearSolver      */
#include <sunnonlinsol/sunnonlinsol_fixedpoint.h> /* access to the fixed point SUNNonlinearSolver */
#include <sundials/sundials_types.h>              /* definition of realtype                       */
#include "SUNDIALSflags55.h"
#include "SUNDIALSinterface55.h"

/* Definitions for IDA */
#define ABSTOL  1.0e-6  
#define RELTOL  1.0e-6  
#define MAXNUMSTEPS     100000  /* default in Sundials is 500 but higher is better for general purpose */
#define MAXERRTESTFAILS 50      /* default is 7 */
#define MAXNUMSTEPSIC   5000  /* default in Sundials is 5 but higher is better for general purpose */
#define MAXNUMJACSIC    5000  /* default in Sundials is 4 but higher is better for general purpose */
#define MAXNUMITERSIC   10000  /* default in Sundials is 10 but higher is better for general purpose */
#define MAXNUMBACKSIC   100000  /* default in Sundials is 100 but higher is better for general purpose */

typedef struct{
    SUNDIALSObject sundialsobject;
    // Integrator setup
    void *integrator_mem;
    SUNMatrix A;
    SUNLinearSolver LS;
    N_Vector u;
    N_Vector udot;
    N_Vector id;
    int *eventstatus;
    // User setup
    void *user_data;
    int nrstates;
    int nrevents;
    int nrfeatures;
    // Simulations options
    int showintegratorstats;
    int maxstep;
    long int maxnumsteps;
    double reltol;
    double abstol;
    int maxerrtestfails;
    int maxorder;
    int maxconvfails;
    double initstep;
    int maxnonliniters;
    int calcic;
    int maxnumstepsic;
    int maxnumjacsic;
    int maxnumitersic;
    int maxnumbacksic;
}IDAObject;

/* Internal functions */
static int IDA55_setOptions(void *self, PyObject *options);
static int IDA55_integrate(void *self, int nrtimesteps, double *timevector, double *statevector, double *derivativevector);
static void IDA55_free(void *self);
static void handleIDAOptions(IDAObject*, PyObject*);
static void createIDAIntegratorMemory(IDAObject *);
static void setIDAIntegratorOptions(IDAObject *);
static void reportIDAStats(void *integrator_mem);
static void freeIDAIntegratorMemory(IDAObject *);

/* Functions containing the model equations */
extern void model(void *simData, double time_local, double *statevector, double *derivativevector, double *RESvector, double *featurevector, int DOflag, int timeindex, double *eventvector, int *eventstatus);
  
/*
========================================================================
 IDA init functions
========================================================================
*/
SUNDIALSObject *
IDA55_init(void *user_data, int nrstates, int nrevents, int nrfeatures, double *idvector, PyObject *options)
{
    IDAObject *ida;
    // Create SUNDIALS idaory
    ida = (IDAObject *) malloc(sizeof(IDAObject));

    // Simulation options
    handleIDAOptions(ida, options);
    if(PyErr_Occurred()){
        free(ida);
        return NULL;
    }
    // SUNDIALS Object
    ida->sundialsobject.integrate = IDA55_integrate;
    ida->sundialsobject.setoptions = IDA55_setOptions;
    ida->sundialsobject.free = IDA55_free;
    ida->sundialsobject.simdata.nreventstriggered = 0;
    ida->sundialsobject.simdata.eventstatusdata = NULL;
    ida->sundialsobject.simdata.eventtimedata = NULL;
    // User setup
    ida->user_data = user_data;
    ida->nrstates = nrstates;
    ida->nrevents = nrevents;
    ida->nrfeatures = nrfeatures;
    // Integrator setup
    ida->u = N_VMake_Serial(nrstates,NULL);
    ida->udot = N_VMake_Serial(nrstates,NULL);
    ida->id = N_VMake_Serial(nrstates,idvector);
    ida->eventstatus = (int *) calloc(nrevents, sizeof(int));
    ida->A = SUNDenseMatrix(nrstates, nrstates);
    ida->LS = SUNLinSol_Dense(ida->u, ida->A);

    createIDAIntegratorMemory(ida);
    setIDAIntegratorOptions(ida);
    
    return (SUNDIALSObject *) ida; 
}

/*
========================================================================
 IDA set options
========================================================================
*/
static int IDA55_setOptions(void *self, PyObject *options)
{
    IDAObject *ida = (IDAObject *) self;
    handleIDAOptions(ida, options);
    if(PyErr_Occurred())
        return -1;
    setIDAIntegratorOptions(ida);
    return 0;
}

/*
========================================================================
 IDA integrate
========================================================================
*/
static int IDA55_integrate(void *self, int nrtimesteps, double *timevector, double *statevector, double *derivativevector)
{
    int k, flag, timeindex, calcic;
    double treturn, tendstep, *eventtimedata, *featuredata;
    int nrevents, nreventstriggered, *eventstatus, nrfeatures;
    char *eventstatusdata;
    void *integrator_mem, *user_data;
    N_Vector u, udot;
    IDAObject *ida;

    if(nrtimesteps <= 1){
        PyErr_SetString(PyExc_ValueError, "'timevector' input argument needs to be a vector if you want to do a simulation.");
        return -1;
    }

    ida = (IDAObject *) self;
    integrator_mem = ida->integrator_mem;
    user_data = ida->user_data;
    eventstatus = ida->eventstatus;
    nrevents = ida->nrevents;
    nrfeatures = ida->nrfeatures;
    calcic = ida->calcic;
    u = ida->u;
    udot = ida->udot;
    N_VSetArrayPointer(statevector, u);
    N_VSetArrayPointer(derivativevector, udot);

    IDAReInit(integrator_mem, timevector[0], u, udot);
    // Calc IC
    if(calcic){
        flag = IDACalcIC(integrator_mem, IDA_YA_YDP_INIT, timevector[1]);
        if(flag < 0){
            PyErr_SetString(PyExc_RuntimeError, "IDA Error: Could not calculate initial conditions");
            return -1;
        }
    }
    /* Simulations data */
    nreventstriggered = 0;
    eventstatusdata = NULL;
    eventtimedata = NULL;
    featuredata = (double *) malloc(nrtimesteps * nrfeatures * sizeof(double));

    // calculate feaures fore time point 0
    model(user_data, timevector[0], statevector, derivativevector, NULL, featuredata, DOFLAG_FEATURE, 0, NULL, NULL);
    // Simulate
    timeindex = 1;
    tendstep = timevector[1]; 
    while(1) {
        flag = IDASolve(integrator_mem, tendstep, &treturn, u, udot, IDA_NORMAL);
        if (flag < 0) {
            if (flag == IDA_TOO_MUCH_WORK)
                PyErr_SetString(PyExc_RuntimeError, "IDA Error: IDA_TOO_MUCH_WORK");
            else if (flag == IDA_TOO_MUCH_ACC)
                PyErr_SetString(PyExc_RuntimeError, "IDA Error: IDA_TOO_MUCH_ACC");
            else if (flag == IDA_ERR_FAIL || flag == IDA_CONV_FAIL)
                PyErr_SetString(PyExc_RuntimeError, "IDA Error: IDA_ERR_FAILURE");
            else
                PyErr_SetObject(PyExc_RuntimeError, PyUnicode_FromFormat("IDA Error Flag: %d", flag));
            free(eventstatusdata);
            free(eventtimedata);
            free(featuredata);
            return -1;
        }
        if (flag == IDA_ROOT_RETURN) {
            /* Event happened */
            IDAGetRootInfo(integrator_mem, eventstatus);
            model(user_data, treturn, statevector, derivativevector, NULL, NULL, DOFLAG_EVENTASSIGN, -1, NULL, eventstatus);
            IDAReInit(integrator_mem, treturn, u, udot);
            // Calc IC
            if(calcic){
                if(tendstep != treturn)
                    flag = IDACalcIC(integrator_mem, IDA_YA_YDP_INIT, tendstep);
                else if(timeindex < (nrtimesteps - 1))
                    flag = IDACalcIC(integrator_mem, IDA_YA_YDP_INIT, timevector[timeindex + 1]);
                if(flag < 0){
                    PyErr_SetString(PyExc_RuntimeError, "IDA Error: Could not calculate initial conditions after event");
                    free(eventstatusdata);
                    free(eventtimedata);
                    free(featuredata);
                    return -1;
                }
            }
            nreventstriggered += 1;  
            if (nreventstriggered == 1) {
                eventtimedata = (double *) calloc(1, sizeof(double));
                eventstatusdata = (char *) calloc(nrevents, sizeof(char));
            } else {
                eventtimedata = (double *) realloc((void *) eventtimedata, nreventstriggered * sizeof(double));
                eventstatusdata = (char *) realloc((void *) eventstatusdata, nrevents * nreventstriggered * sizeof(char));
            }
            eventtimedata[nreventstriggered - 1] = treturn;
            for (k = 0; k < nrevents; k ++)
                eventstatusdata[nrevents * (nreventstriggered - 1) + k] = (char) eventstatus[k];
        }
        if (tendstep == treturn) { 
            model(user_data, tendstep, statevector, derivativevector, NULL, &featuredata[timeindex * nrfeatures], DOFLAG_FEATURE, timeindex, NULL, NULL);
            
            timeindex += 1;
            if (timeindex < nrtimesteps) 
                tendstep = timevector[timeindex];
            else
                break;
        }
    }

    ida->sundialsobject.simdata.nreventstriggered = nreventstriggered;
    ida->sundialsobject.simdata.eventtimedata = eventtimedata;
    ida->sundialsobject.simdata.eventstatusdata = eventstatusdata;
    ida->sundialsobject.simdata.featuredata = featuredata;

    if(ida->showintegratorstats){
        reportIDAStats(ida->integrator_mem);
    }

    return 0;
}

/*
========================================================================
 IDA free
========================================================================
*/
static void IDA55_free(void *self)
{
    IDAObject *ida = (IDAObject *) self;
    freeIDAIntegratorMemory(ida);
    N_VDestroy_Serial(ida->u);
    N_VDestroy_Serial(ida->udot);
    N_VDestroy_Serial(ida->id);
    SUNLinSolFree(ida->LS); /* Free the linear solver */
    SUNMatDestroy(ida->A); /* Free the matrix used in linear solver */
    free(ida->eventstatus);
    free(ida);
}

/*
========================================================================
 RES function R(t,u,udot)
========================================================================
*/
static int res(double time, N_Vector u, N_Vector udot, N_Vector res, void *user_data)
{
    double *statevector, *derivativevector, *resvector;
    /* connect input and result data */
    statevector = N_VGetArrayPointer(u);
    derivativevector = N_VGetArrayPointer(udot);
    resvector = N_VGetArrayPointer(res);
    /* run the model */
    model(user_data, time, statevector, derivativevector, resvector, NULL, DOFLAG_RESIDUAL, -1, NULL, NULL);
    return(0);
}

/*
========================================================================
 Event function 
========================================================================
*/
static int g(double time, N_Vector u, N_Vector udot, double *gout, void *user_data)
{
    double *statevec, *derivativevec;
    /* connect input data */
    statevec = N_VGetArrayPointer(u);
    derivativevec = N_VGetArrayPointer(udot);
    /* run the event function */
    model(user_data, time, statevec, derivativevec, NULL, NULL, DOFLAG_EVENT, -1, gout, NULL);
    return(0);
}

/*
========================================================================
 Free integrator idaory 
========================================================================
*/
static void freeIDAIntegratorMemory(IDAObject *ida)
{
    IDAFree(&ida->integrator_mem); /* Free the integrator */
}

/*
==============================================
INITIALIZE INTEGRATOR
==============================================
*/

static void createIDAIntegratorMemory(IDAObject *ida)
{
    int k, *rootdir;
    void *integrator_mem;
    double *vector;

    // creaty memory
    integrator_mem = IDACreate();
    // Init
    vector = (double *) malloc(ida->nrstates * sizeof(double)); // Dummy variable so initialization can be done
    N_VSetArrayPointer(vector, ida->u);
    N_VSetArrayPointer(vector, ida->udot);
    IDAInit(integrator_mem, res, 0, ida->u, ida->udot);

    // Attach solvers
    IDASetLinearSolver(integrator_mem, ida->LS, ida->A);
    IDASetJacFn(integrator_mem, NULL); // Use a difference quotient Jacobian

    // Add event function
    if (ida->nrevents > 0) {
        IDARootInit(integrator_mem, ida->nrevents, g);
        // Set root dir = 1 => increasing direction
        rootdir = (int *) malloc(ida->nrevents * sizeof(int));
        for(k = 0; k < ida->nrevents; k++)
            rootdir[k] = 1;
        IDASetRootDirection(integrator_mem, rootdir);
        free(rootdir);
    }

    ida->integrator_mem = integrator_mem;
    free(vector);
}

static void setIDAIntegratorOptions(IDAObject *ida)
{
    void *integrator_mem;
    integrator_mem = ida->integrator_mem;
    // Set options
    IDASStolerances(integrator_mem, ida->reltol, ida->abstol);
    IDASetId(integrator_mem, ida->id);
    IDASetUserData(integrator_mem, ida->user_data);
    if (ida->maxstep > 0) IDASetMaxStep(integrator_mem, ida->maxstep);
    if (ida->maxnumsteps > 0) IDASetMaxNumSteps(integrator_mem, ida->maxnumsteps);
    IDASetMaxErrTestFails(integrator_mem, ida->maxerrtestfails);
    IDASetMaxOrd(integrator_mem, ida->maxorder);
    IDASetMaxConvFails(integrator_mem, ida->maxconvfails);
    IDASetInitStep(integrator_mem, ida->initstep);
    IDASetMaxNonlinIters(integrator_mem, ida->maxnonliniters);
    // calc IC options
    IDASetMaxNumStepsIC(integrator_mem, ida->maxnumstepsic);
    IDASetMaxNumJacsIC(integrator_mem, ida->maxnumjacsic);
    IDASetMaxNumItersIC(integrator_mem, ida->maxnumitersic);
    IDASetMaxBacksIC(integrator_mem, ida->maxnumbacksic);
}

/*
==============================================
 HANDLE THE SIMULATION OPTIONS
==============================================
*/
static void handleIDAOptions(IDAObject *ida, PyObject *options)
{
    PyObject *item;
    /* check options */
    ida->showintegratorstats         = 0;
    ida->maxstep                     = 0.0;              /* default value in IDA => infinity */
    ida->maxnumsteps                 = MAXNUMSTEPS;      /* we set a value different from IDA default values as default */
    ida->maxerrtestfails             = MAXERRTESTFAILS;  /* we set a value different from IDA default values as default */
    ida->maxorder                    = 5;                /* default value for default method (stiff) */
    ida->maxconvfails                = 10;               /* default value in IDA */
    ida->initstep                    = 0.0;              /* default value in IDA */
    ida->maxnonliniters              = 3;                /* default value in IDA */
    ida->calcic                      = 0;                /* 0 = don't calculate IC */
    ida->reltol                      = RELTOL;           /* we set a value different from IDA default values as default */
    ida->abstol                      = ABSTOL;           /* we set a value different from IDA default values as default */
    ida->maxnumstepsic               = MAXNUMSTEPSIC;      /* we set a value different from IDA default values as default */
    ida->maxnumjacsic                = MAXNUMJACSIC;      /* we set a value different from IDA default values as default */
    ida->maxnumitersic               = MAXNUMITERSIC;      /* we set a value different from IDA default values as default */
    ida->maxnumbacksic               = MAXNUMBACKSIC;      /* we set a value different from IDA default values as default */
    if(PyDict_Check(options)){
        if (PyDict_Size(options) != 0) {
            item = PyDict_GetItemString(options, "showintegratorstats");
            if (item != NULL) {
                if (PyObject_IsTrue(item))
                    ida->showintegratorstats = 1; 
                else 
                    ida->showintegratorstats = 0;
            }
            item = PyDict_GetItemString(options, "maxstep");
            if (item != NULL) {
                if (PyFloat_Check(item)) {
                    ida->maxstep = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxstep' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "maxnumsteps");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    ida->maxnumsteps = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxnumsteps' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "calcic");
            if (item != NULL) {
                if (PyObject_IsTrue(item))
                    ida->calcic = 1;
                else 
                    ida->calcic = 0;
            }
            item = PyDict_GetItemString(options, "reltol");
            if(item != NULL) {
                if (PyFloat_Check(item)) {
                    ida->reltol = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.reltol' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "abstol");
            if (item != NULL) {
                if (PyFloat_Check(item)) {
                    ida->abstol = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.abstol' wrongly defined.");
                }
            }
            /* Additional Options */
            item = PyDict_GetItemString(options, "maxerrtestfails");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    ida->maxerrtestfails = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxerrtestfails' wrongly defined.");         
                }
            }   

            item = PyDict_GetItemString(options, "maxorder");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    ida->maxorder = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxorder' wrongly defined.");    
                }
            }        

            item = PyDict_GetItemString(options, "maxconvfails");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    ida->maxconvfails = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxconvfails' wrongly defined."); 
                }
            }   
            
            item = PyDict_GetItemString(options, "initstep");
            if (item != NULL) {
                if (PyFloat_Check(item)) {
                    ida->initstep = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.initstep' wrongly defined.");                
                }
            }
            item = PyDict_GetItemString(options, "maxnonliniters");
            if (item != NULL) {
                if (PyFloat_Check(item)) {
                    ida->maxnonliniters = PyFloat_AsDouble(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxnonliniters' wrongly defined.");  
                }
            }
            item = PyDict_GetItemString(options, "maxnumstepsic");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    ida->maxnumstepsic = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxnumstepsic' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "maxnumjacsic");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    ida->maxnumjacsic = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxnumjacsic' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "maxnumitersic");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    ida->maxnumitersic = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxnumitersic' wrongly defined.");
                }
            }
            item = PyDict_GetItemString(options, "maxnumbacksic");
            if (item != NULL) {
                if (PyLong_Check(item)) {
                    ida->maxnumbacksic = PyLong_AsLong(item); 
                }
                else {
                    PyErr_SetString(PyExc_ValueError, "'options.maxnumbacksic' wrongly defined.");
                }
            }       
        }
    } else
        PyErr_SetString(PyExc_ValueError, "Options needs to be a dictionary.");  
    return;
}

/*
========================================================================
 Integration statistics report function
========================================================================
*/
static void reportIDAStats(void *integrator_mem)
{
    long int STATS_nsteps;
    long int STATS_nfevals;
    long int STATS_netfails;
    double STATS_hinused;
    double STATS_tolsfac;
    /* Get the statistics */
    IDAGetNumSteps(integrator_mem,&STATS_nsteps);
    IDAGetNumResEvals(integrator_mem,&STATS_nfevals);
    IDAGetNumErrTestFails(integrator_mem,&STATS_netfails);
    IDAGetActualInitStep(integrator_mem,&STATS_hinused);
    IDAGetTolScaleFactor(integrator_mem,&STATS_tolsfac);
    /* Report statistics */
    printf("\nIntegrator Statistics\n");
    printf("=====================\n");
    printf("Cumulative number of internal steps:    %ld\n",STATS_nsteps);
    printf("No. of calls to r.h.s. function:        %ld\n",STATS_nfevals);
    printf("No. of local error test failures:       %ld\n",STATS_netfails);
    printf("Actual init step size used:             %g\n",STATS_hinused);
    printf("Suggested factor for tolerance scaling: %g\n\n",STATS_tolsfac);
}
