import os
from setuptools import sandbox
import re
from collections import OrderedDict as Odict
import sys

_MeDigiTFolder = os.path.dirname(os.path.realpath(__file__))

class _HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout

##################################################################################################
# Compile new model
##################################################################################################
def _modelStructure(file):
	# Read in file and compile
	##############################################################################################
	try:
		fileHandle = open(file,'r')
		fileTxt = fileHandle.read()
		fileHandle.close()

		sectionNames = ["NAME", "METADATA", "MACROS", "STATES", "PARAMETERS", "VARIABLES",
						"FUNCTIONS", "EVENTS", "OUTPUTS", "INPUTS", "FEATURES"]

		sectionTxt = _txtParts(fileTxt,sectionNames)
		modelStructure = dict.fromkeys(["FILE"] + sectionNames,None)
		modelStructure["FILE"] = file
		modelStructure["NAME"] = _removeWhiteSpace(sectionTxt["NAME"])
		modelStructure["METADATA"] = _compileMetadata(sectionTxt["METADATA"])
		modelStructure["MACROS"] = _compileMacros(sectionTxt["MACROS"])
		states,algequations = _compileStateEquations(sectionTxt["STATES"])
		modelStructure["STATES"] = states
		modelStructure["ALGEQUATIONS"] = algequations
		modelStructure["PARAMETERS"] = _compileParameters(sectionTxt["PARAMETERS"])
		modelStructure["FUNCTIONS"] = _compileFunctions(sectionTxt["FUNCTIONS"])
		modelStructure["EVENTS"] = _compileEvents(modelStructure,sectionTxt["EVENTS"])
		modelStructure["OUTPUTS"] = _compileOutputs(sectionTxt["OUTPUTS"])
		modelStructure["INPUTS"] = _compileInputs(sectionTxt["INPUTS"])
		modelStructure["VARIABLES"] = _compileVariables(modelStructure,sectionTxt["VARIABLES"])
		modelStructure["FEATURES"] = _compileFeatures(sectionTxt["FEATURES"])

		return modelStructure
	except Exception as e:
		raise ValueError(f'Error while compiling txt-file: {file}') from e

def _modelCFile(model):
	try:
		# Create .c file
		##############################################################################################
		filedir = os.path.dirname(model["FILE"])
		cfile = os.path.join(filedir,model["NAME"] + ".c")
		CfileHandle = open(cfile,'w')
		_printModelToFile(CfileHandle,model)
		CfileHandle.close()
		return cfile
		
	except Exception as e:
		raise ValueError(f"Error while compiling model-structure: '{model['NAME']}'") from e

def _modelModuleFile(cfile):
	try:
		with _HiddenPrints():
			sandbox.run_setup(os.path.join(_MeDigiTFolder,'_setupModel_.py'), ['build_ext', '-b', os.path.dirname(_MeDigiTFolder), 'clean', cfile])
	except Exception as e:
		raise ValueError(f"Error while compiling c-file: {cfile}") from e
	
# Write model to .c file
##############################################################################################
def _printModelToFile(file,model):
	file.write("#define MeDigiT_MODEL\n")
	file.write("#include <Python.h>\n")
	file.write("#include <math.h>\n")
	file.write('#include "mathaddon.h"\n')
	file.write('#include "kineticformulas.h"\n')
	file.write('#include "SUNDIALSflags55.h"\n')
	file.write('#include "_Models.h"\n\n')

	for macro,itemdict in model["MACROS"].items():
		file.write(f'#define {macro} {itemdict["expression"]}\n')

	file.write('\nstatic ModelFunction {}_modelfunction;\n'.format(model["NAME"]))
	file.write('static ICFunction {}_initialcondition;\n'.format(model["NAME"]))
	file.write('static PyObject *{}_new(PyTypeObject *type, PyObject *args, PyObject *kwds);\n\n'.format(model["NAME"]))

	# declare names
	file.write('// states, features, outputs, inputs, events, parameters\n')
	_listDeclaration(file, model["NAME"] + "_numberof", [len(model["STATES"]),len(model["FEATURES"]),len(model["OUTPUTS"]),len(model["INPUTS"]),len(model["EVENTS"]),len(model["PARAMETERS"])], type=1)
	_listDeclaration(file, model["NAME"] + "_statenames", list(model["STATES"].keys()), type=0)
	_listDeclaration(file, model["NAME"] + "_featurenames", list(model["FEATURES"].keys()), type=0)
	_listDeclaration(file, model["NAME"] + "_featureunits", [itemdict["unit"] for _,itemdict in model["FEATURES"].items()], type=0)
	_listDeclaration(file, model["NAME"] + "_outputnames", list(model["OUTPUTS"].keys()), type=0)
	_listDeclaration(file, model["NAME"] + "_inputnames", [itemdict["externvariable"] for _,itemdict in model["INPUTS"].items()], type=0)
	_listDeclaration(file, model["NAME"] + "_parameternames", list(model["PARAMETERS"].keys()), type=0)
	_listDeclaration(file, model["NAME"] + "_eventnames", list(model["EVENTS"].keys()), type=0)

	# decalare mandatory inputs, i.e. inputs without default value
	file.write('\n// {0 = mandatory, 1 = non-mandatory, 2 = non-mandatory - zero default value}\n')
	mandatoryinputs = [0]*len(model["INPUTS"]) # 0 mandatory, 1 non-mandatory, 2 non-mandatory - default value 0
	for i,itemdict in enumerate(model["INPUTS"].values()):
		if itemdict["defaultvalue"]:
			if itemdict["defaultvalue"] == "ZERO":
				mandatoryinputs[i] = 2
			else:
				mandatoryinputs[i] = 1
	_listDeclaration(file, model["NAME"] + "_mandatoryinputs", mandatoryinputs, type=1)

	# declare algebraic states, i.e. a state whose derivative does not appear in state expressions
	file.write('// {#nrdifferentialstates,<differentialstate index>}\n')
	differentialstates = [i for i,state in enumerate(model["STATES"].values()) if not state["isalgebraic"]]
	_listDeclaration(file, model["NAME"] + "_differentialstates", [len(differentialstates)] + differentialstates, type=1)

	# declare dependency on input, i.e. which inputs are needed to compute outputs
	file.write('// {#depency,<input numbers>}\n')
	inputdependency = _getInputDependency(model)
	_listDeclaration(file, model["NAME"] + "_inputdependency", [len(inputdependency)] + inputdependency, type=1)

	# default parameters
	_listDeclaration(file, model["NAME"] + "_defaultparameters", [itemdict["expression"] for _,itemdict in model["PARAMETERS"].items()], type=2)
	file.write('const char {}_timeunit[] = "{}";\n\n'.format(model["NAME"], model["METADATA"]["Timeunit"]))

	for funcname,itemdict in model["FUNCTIONS"].items():
		file.write('static double {}('.format(funcname))
		args = itemdict["args"]
		for i in range(len(args)):
			if i < len(args) - 1:
				file.write('double {},'.format(args[i]))
			else:
				file.write('double {}'.format(args[i]))
		file.write('){{\n\treturn {};\n'.format(itemdict["expression"]))
		file.write('}\n\n')

	# Model function
	file.write('static void {}_modelfunction(double time, double timescale, double *statevector, double *derivativevector, double *RESvector, double *parametervector, double *featurevector, double *outputvector, double **inputvector, double *eventvector, int *eventstatus, int DOflag){{\n'.format(model["NAME"]))
	_variableDeclaration(file,list(model["STATES"].keys()))
	if model["ALGEQUATIONS"]: 
		# model is a DAE, derivatives can be used as any other variable, defined as ddt_state
		file.write('\tdouble timescale_inv = 1 / timescale;\n') # needed to convert to local derivativevector
		_variableDeclaration(file,["ddt_" + name for name,state in model["STATES"].items() if not state["isalgebraic"]])
	_variableDeclaration(file,list(model["PARAMETERS"].keys()))
	_variableDeclaration(file,list(model["INPUTS"].keys()))
	_variableDeclaration(file,list(model["VARIABLES"].keys()))
	file.write('\n')

	# states
	for i,name in enumerate(model["STATES"]):
		file.write('\t{} = statevector[{}];\n'.format(name,i))
	# derivatives	
	if model["ALGEQUATIONS"]:
		for i,(name,state) in enumerate(model["STATES"].items()): 
			if not state["isalgebraic"]: # No derivative for algebraic states
				file.write('\t{} = timescale_inv * derivativevector[{}];\n'.format("ddt_" + name, i))
	# parameters
	for i,name in enumerate(model["PARAMETERS"]):
		file.write('\t{} = parametervector[{}];\n'.format(name,i))
	# inputs
	for i,(name,itemdict) in enumerate(model["INPUTS"].items()):
		expression = re.sub(r'\b[a-zA-Z_]\w*\b(?!\()',f'(*inputvector[{i}])',itemdict["expression"])
		if itemdict["defaultvalue"] and itemdict["defaultvalue"] != "ZERO":
			file.write(f'\t{name} = (inputvector[{i}]) ? {expression} : {itemdict["defaultvalue"]};\n') # non-mandatory inputs
		else:
			file.write(f'\t{name} = {expression};\n') # mandatory inputs
	# variables
	for name,itemdict in model["VARIABLES"].items():
		file.write('\t{} = {};\n'.format(name,itemdict["expression"]))
	
	# DDT
	file.write('\n\tif (DOflag == DOFLAG_DDT) {\n')
	# only for ODE 
	if not model["ALGEQUATIONS"]:
		for i,(_,itemdict) in enumerate(model["STATES"].items()):
			file.write('\t\tderivativevector[{}] = timescale * ({});\n'.format(i,itemdict["expression"]))
	file.write('\t}')

	# OUTPUT
	file.write(' else if (DOflag == DOFLAG_OUTPUT) {\n')
	for i,(_,itemdict) in enumerate(model["OUTPUTS"].items()):
		file.write('\t\toutputvector[{}] = {};\n'.format(i,itemdict["expression"]))
	file.write('\t}')

	# FEATURE
	file.write(' else if (DOflag == DOFLAG_FEATURE) {\n')
	for i,(_,itemdict) in enumerate(model["FEATURES"].items()):
		file.write('\t\tfeaturevector[{}] = {};\n'.format(i,itemdict["expression"]))
	file.write('\t}')

	# EVENT
	file.write(' else if (DOflag == DOFLAG_EVENT) {\n')
	for i,(_,itemdict) in enumerate(model["EVENTS"].items()):
		file.write('\t\teventvector[{}] = ({}) - 0.5;\n'.format(i,itemdict["condition"]))
	file.write('\t}')

	# EVENTASSIGN
	file.write(' else if (DOflag == DOFLAG_EVENTASSIGN) {\n')
	for i,(_,itemdict) in enumerate(model["EVENTS"].items()):
		file.write(f'\t\tif(eventstatus[{i}] == 1){{\n')
		for assign in itemdict["assignments"]:
			if assign["vartype"] == 0: # state
				file.write(f'\t\t\tstatevector[{assign["varindex"]}] = {assign["expression"]};\n')
			elif assign["vartype"] == 1: # parameter
				file.write(f'\t\t\tparametervector[{assign["varindex"]}] = {assign["expression"]};\n')
		file.write('\t\t}\n')
	file.write('\t}')

	# RESIDUAL
	file.write(' else if (DOflag == DOFLAG_RESIDUAL) {\n')
	# ODE:s
	ODEstates = [(i,state) for i,state in enumerate(model["STATES"].values()) if "expression" in state]
	for k in range(len(ODEstates)):
		if k != ODEstates[k][0]: # here as a check, ODE states should be at the top of model["STATES"]
			raise ValueError("programming error...bug in code")
		file.write('\t\tRESvector[{}] = timescale * ({}) - derivativevector[{}];\n'.format(k,ODEstates[k][1]["expression"],k)) 
	# Algebraic
	k = len(ODEstates)
	for alg in model["ALGEQUATIONS"]:
		file.write('\t\tRESvector[{}] = timescale * ({});\n'.format(k,alg))
		k += 1
	file.write('\t}\n')

	file.write('}\n\n')

	# IC function
	file.write('/* Function for initial condition calculation */\n')
	file.write(f'static void {model["NAME"]}_initialcondition(double *icvector, double *dericvector, double *parametervector){{\n')
	
	_variableDeclaration(file,list(model["PARAMETERS"].keys()))
	_variableDeclaration(file,[variable for variable,itemdict in model["VARIABLES"].items() if not itemdict["inputdependency"] and not itemdict["statedependent"]])

	for i,name in enumerate(model["PARAMETERS"]):
		file.write('\t{} = parametervector[{}];\n'.format(name,i))
	for name,itemdict in ((name,itemdict) for name,itemdict in model["VARIABLES"].items() if not itemdict["inputdependency"] and not itemdict["statedependent"]):
		file.write('\t{} = {};\n'.format(name,itemdict["expression"]))
	for i,(_,itemdict) in enumerate(model["STATES"].items()):
		file.write('\ticvector[{}] = {};\n'.format(i,itemdict["ic"]))
		file.write('\tdericvector[{}] = {};\n'.format(i,itemdict["deric"]))
	
	file.write('}\n\n')

	file.write('/* Functions for Python C API Extension */\n')
	file.write(f'static PyTypeObject {model["NAME"]}_Type = {{\n')
	file.write('\tPyVarObject_HEAD_INIT(NULL, 0)\n')
	file.write(f'\t.tp_name =  "MeDigiT.Models.{model["NAME"]}.{model["NAME"]}",\n')
	file.write('\t.tp_doc = "{} model",\n'.format(model["NAME"]))
	file.write('\t.tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,\n')
	file.write(f'\t.tp_new = {model["NAME"]}_new,\n')
	file.write('};\n\n')

	file.write(f'static ModelStructure {model["NAME"]} = {{\n')
	file.write(f'\t.function = {model["NAME"]}_modelfunction,\n')
	file.write(f'\t.initialcondition = {model["NAME"]}_initialcondition,\n')
	file.write(f'\t.numberof = {model["NAME"]}_numberof,\n')
	file.write(f'\t.statenames = {model["NAME"]}_statenames,\n')
	file.write(f'\t.featurenames = {model["NAME"]}_featurenames,\n')
	file.write(f'\t.featureunits = {model["NAME"]}_featureunits,\n')
	file.write(f'\t.outputnames = {model["NAME"]}_outputnames,\n')
	file.write(f'\t.inputnames = {model["NAME"]}_inputnames,\n')
	file.write(f'\t.parameternames = {model["NAME"]}_parameternames,\n')
	file.write(f'\t.eventnames = {model["NAME"]}_eventnames,\n')
	file.write(f'\t.differentialstates = {model["NAME"]}_differentialstates,\n')
	file.write(f'\t.inputdependency = {model["NAME"]}_inputdependency,\n')
	file.write(f'\t.defaultparameters = {model["NAME"]}_defaultparameters,\n')
	file.write(f'\t.timeunit = {model["NAME"]}_timeunit,\n')
	file.write(f'\t.has_algebraic_eq = {1 if model["ALGEQUATIONS"] else 0},\n')
	file.write(f'\t.mandatoryinputs = {model["NAME"]}_mandatoryinputs\n')
	file.write('};\n\n')

	file.write('static PyObject *\n')
	file.write(f'{model["NAME"]}_new(PyTypeObject *type, PyObject *args, PyObject *kwds){{\n')
	file.write('\tModelObject *self;\n')
	file.write('\tself = (ModelObject *) type->tp_alloc(type, 0);\n')
	file.write('\tif (self) {\n')
	file.write(f'\t\tself->model = &{model["NAME"]};\n')
	file.write('\t\tif(Model_alloc(self) < 0){\n')
	file.write('\t\t\tPy_DECREF(self);\n')
	file.write('\t\t\treturn NULL;\n')
	file.write('\t\t}\n')
	file.write('\t}\n')
	file.write('\treturn (PyObject *) self;\n')
	file.write('}\n\n')

	file.write(f'static PyModuleDef {model["NAME"]}Module = {{\n')
	file.write('\tPyModuleDef_HEAD_INIT,\n')
	file.write(f'\t.m_name = "MeDigiT.Models.{model["NAME"]}",\n')
	file.write(f'\t.m_doc = "{model["NAME"]} Module",\n')
	file.write(f'\t.m_size = -1\n')
	file.write('};\n\n')

	file.write('PyMODINIT_FUNC\n')
	file.write(f'PyInit_{model["NAME"]}(void){{\n')
	file.write('\tPyObject *m;\n\n')

	file.write(f'\tm = PyModule_Create(&{model["NAME"]}Module);\n')
	file.write('\tif (m == NULL)\n')
	file.write('\t\treturn NULL;\n\n')

	file.write('\timport_Models();\n')
	file.write(f'\t{model["NAME"]}_Type.tp_base = Model_Base_Type;\n')
	file.write(f'\tif (PyType_Ready(&{model["NAME"]}_Type) < 0){{\n')
	file.write(f'\t\tPy_DECREF(m);\n')
	file.write('\t\treturn NULL;\n')
	file.write('\t}\n\n')

	file.write(f'\tPy_INCREF(&{model["NAME"]}_Type);\n')

	file.write(f'\tif (PyModule_AddObject(m, "{model["NAME"]}", (PyObject *) &{model["NAME"]}_Type) < 0){{\n')
	file.write(f'\t\tPy_DECREF(&{model["NAME"]}_Type);\n')
	file.write('\t\tPy_DECREF(m);\n')
	file.write('\t\treturn NULL;\n')
	file.write('\t}\n\n')

	file.write('\timport_Models();\n')
	file.write('\treturn m;\n')
	file.write('}')

##################################################################################################
# Compile functions
##################################################################################################

def _compileMacros(txt):
	names,values = _nameValuePair(txt)
	return Odict([(name, {"expression" : _replaceExpr(value)}) for name,value in zip(names,values)])

def _compileParameters(txt):
	names,values = _nameValuePair(txt)
	for name,value in zip(names,values):
		if not _isNumber(_replaceExpr(value)):
			raise ValueError(f"Parameters can only be assigned to numbers, incorrect definition for parameter '{name}': '{value}'")
	return Odict([(name, {"expression" : _replaceExpr(value)}) for name,value in zip(names,values)])

def _compileVariables(model,txt):
	names,values = _nameValuePair(txt)
	variables = Odict([(name, {"expression" : _replaceExpr(value), "inputdependency" : None, "statedependent" : False}) for name,value in zip(names,values)])
	_getVariableDependency(model,variables)
	return variables

def _compileInputs(txt):
	names,values = _nameValuePair(txt)
	extvars = len(names)*[None]
	defvals = len(names)*[None] # default values

	for i,value in enumerate(values):
		k = value.find("@") # mark default value with @
		if k >= 0:
			defvals[i] = _replaceExpr(value[k+1:])
			if defvals[i] == "":
				defvals[i] = "ZERO"
			else:
				try:
					f = float(defvals[i])
					if f == 0.0: # check if default value is zero
						defvals[i] = "ZERO"
				except ValueError:
					pass

			value = value[:k]
			values[i] = value

		varinexpr = list(set(re.findall(r'\b[a-zA-Z_]\w*\b(?!\()',value)))
		if len(varinexpr) != 1:
			raise ValueError(f'Incorrect input definition for input {names[i]}: "{value}"')
		else:
			extvars[i] = varinexpr[0]
	return Odict([(name, {"expression" : _replaceExpr(value), "externvariable" : extvar, "defaultvalue" : defval}) for name,value,extvar,defval in zip(names,values,extvars,defvals)])

def _compileMetadata(txt):
	mandatoryName = ["Timeunit"]
	names,values = _nameValuePair(txt)
	names = [name.title() for name in names]

	for mandatory in mandatoryName:
		if mandatory not in names:
			raise ValueError("Metadata is missing mandatory definition {}.".format(mandatory))
	# Check timeunit
	if not values[names.index("Timeunit")] in ["ns", "μs", "ms", "s", "m", "h", "d", "y"]:
		raise ValueError("Incorrect timeunit definition {}.".format(values[names.index("Timeunit")]))
	return dict(zip(names,values))

def _compileStateEquations(txt):
	names,values = _nameValuePair(txt)
	# change old writing: d/dt(state) to new: ddt_state
	names = [_replaceddt(name) for name in names]
	# check names
	for name in names:
		# look for: derivative expression | algebraic expression | state IC | state derivative IC
		if not re.search(r"^(ddt_\w+|0|\w+\(0\)|ddt_\w+\(0\))$",name):
			raise ValueError(f"Incorrect definition in state section: '{name}'")
	# search for ODE states, here "expression" in the inner dict is the derivative expression, i.e. RHS of d/dty = f(t,y)
	states = Odict([(state[0].group("state"), {"expression" : _replaceExpr(state[1]), "ic" : 0, "deric" : 0, "isalgebraic" : False}) for state in ((re.search(r"^ddt_(?P<state>\w+)$", name), value) for name,value in zip(names,values)) if state[0]])
	# search for algebraic equations
	algebraiceqs = [_replaceExpr(value) for name,value in zip(names,values) if name == "0"]
	nrequations = len(states) + len(algebraiceqs)
	# search for state IC
	ics = dict([(state[0].group("state"), _replaceExpr(state[1])) for state in ((re.search(r"^(?!ddt_)(?P<state>\w+)\(0\)$", name), value) for name,value in zip(names,values)) if state[0]])
	# search for state derivative IC
	derics = dict([(state[0].group("state"), _replaceExpr(state[1])) for state in ((re.search(r"^ddt_(?P<state>\w+)\(0\)$", name), value) for name,value in zip(names,values)) if state[0]])
	# add ODE state IC:s
	for state,itemdict in states.items():
		if state in ics:
			itemdict["ic"] = ics[state]
		else:
			raise ValueError(f"Missing initial condition for ODE state '{state}'")
		if state in derics: # Optional for ODE state
			itemdict["deric"] = derics[state]	
	# Check type alebraic equations present
	if algebraiceqs: # DAE
		# add differential states, defined by having a IC, ddt_state(0), defined for the derivative
		for state,deric in derics.items():
			if state not in states:
				if state not in ics: # diffstates also need a IC for the actual state: state(0)
					raise ValueError(f"Missing initial condition for differential state '{state}'")
				if not _containsDerivate(state,values): # check that derivate is present for differential state
					raise ValueError(f"Missing derivative for differential state '{state}' in state equations")
				states[state] = {"ic" : ics[state], "deric" : deric, "isalgebraic" : False}
		# add algebraic states
		for state,ic in ics.items():
			if state not in states:
				if _containsDerivate(state,values): # check that derivate is not present for algebraic state
					raise ValueError(f"Derivative for algebraic state '{state}' present in state equations")
				states[state] = {"ic" : ic, "deric" : 0, "isalgebraic" : True} # even though algebraic state have no derivative, IDA need a IC for its derivative, i.e. 'deric'
	# check equal number of states as equations
	nrstates = len(states)
	if nrstates != nrequations:
		raise ValueError(f"Unequal number of states: {nrstates} and equations: {nrequations}")
	return states,algebraiceqs


def _compileFunctions(txt):
	names,values = _nameValuePair(txt)
	functions = Odict()
	for name,value in zip(names,values):
		gStart,gEnd = _findFirstGroup(name)
		args = _splitExpr(name[gStart:gEnd],',')
		functions[name[:gStart-1]] =  {"expression" : _replaceExpr(value), "args" : args}
	return functions

def _compileEvents(model,txt):
	statenames = list(model["STATES"].keys())
	parameternames = list(model["PARAMETERS"].keys())

	names,values = _nameValuePair(txt)
	events = Odict()
	for name,value in zip(names,values):
		elements = _splitExpr(value,',')
		if (len(elements) % 2) == 0:
			raise ValueError("Incorrect event definition '{}'".format(value))
		assignments = []
		i = 0
		while(i < len(elements[1:])):
			assign = {"vartype" : None, "varindex" : None, "expression" : None} # vartype: 0 = STATE, 1 = PARAMETER
			var = _removeWhiteSpace(elements[1+i])
			expr = _replaceExpr(elements[2+i])

			if var in statenames:
				assign["vartype"] = 0
				assign["varindex"] = statenames.index(var)
				assign["expression"] = expr
			elif var in parameternames:
				assign["vartype"] = 1
				assign["varindex"] = parameternames.index(var)
				assign["expression"] = expr
			else:
				raise ValueError(f"Incorrect definition of event assign to variable '{var}' which needs to be a parameter or state")
			
			assignments.append(assign)
			i += 2

		events[name] = {"condition" : _replaceExpr(elements[0]), "assignments" : assignments}
	return events

def _compileOutputs(txt):
	names,values = _nameValuePair(txt)
	outputs = Odict([(name, {"expression" : _replaceExpr(value)}) for name,value in zip(names,values)])
	return outputs

def _compileFeatures(txt):
	names,values = _nameValuePair(txt, keepspace = True)
	features = Odict()

	for name,value in zip(names,values):
		i = value.find("#")
		if i < 0:
			unit = "1"
		else:
			unit = value[i+1:]
			value = value[:i]
		features[name] = {"expression" : _replaceExpr(value), "unit" : _removeWhiteSpace(unit)}
	return features

##################################################################################################
# Helper functions
##################################################################################################
def _isNumber(s):
    try:
    	float(s)
    	return True
    except ValueError:
        return False

def _findFile(name, path):
    for root, dirs, files in os.walk(path):
        if name in files:
            return os.path.join(root, name)
    return ''

def _txtParts(txt,sectionNames):
	# remove comments
	txt = _removeComments(txt)

	# find section parts
	indices = []
	for section in sectionNames:
		indices.append({'Section': section, 'Span' : _sectionSpan(txt, section)})
	indices.append({'Section' : 'END', 'Span' : (len(txt) + 1,-1)})
	indices = sorted(indices, key = lambda x : x['Span'][0])

	section = dict()
	for i in range(len(indices) - 1):
		section[indices[i]['Section']] = txt[indices[i]['Span'][1]:indices[i+1]['Span'][0]-1].strip()
	return section

def _sectionSpan(txt, section):
	s = re.search('########## {}'.format(section),txt)
	if not s:
		raise ValueError('Textfile missing {}-section.'.format(section))
	return s.span()

def _removeWhiteSpace(txt):
	return re.sub(r'\s+','',txt)

def _nameValuePair(txt, keepspace = False):
	indices = _declarationIndices(txt)
	indices.append(len(txt))
	names = []
	values = []
	for i in range(len(indices) - 1):
		txtLine = txt[indices[i]:indices[i+1]]
		equalIndex = txtLine.find("=")
		if(keepspace):
			names.append(txtLine[0:equalIndex].strip())
		else:
			names.append(_removeWhiteSpace(txtLine[0:equalIndex]))
		values.append(txtLine[equalIndex+1:].strip())
	return names,values

def _declarationIndices(txt):
	indices = []
	for m in re.finditer(r'\n[^\n]*=', '\n' + txt):
		indices.append(m.start())
	return indices

def _convertToDouble(txt):
	return re.sub(r'\b(?<!\.)(?<![eE][-+])(\d+)(?!\.)(\b|[eE][\d+-])',r'\g<1>.0\g<2>',txt)

def _convertPowerExpr(txt):
	txtorg = txt
	while True:
		try:
			i = txt.index('^')
		except Exception:
			break
		preExpr = txt[:i]
		postExpr = txt[i+1:]
		try:
			baseStart = _findBaseStart(preExpr)
			expEnd = _findExpEnd(postExpr)
		except Exception:
			raise ValueError("Incorrect equation definition: '{}'".format(txtorg)) from None
		
		txt = "{}pow({},{}){}".format(preExpr[:baseStart],preExpr[baseStart:],postExpr[:expEnd],postExpr[expEnd:])
	return txt

def _findBaseStart(expr):
	m = re.search(r'(\d+(\.\d+)?|\w+)$',expr)
	if m:
		return len(expr)-len(m.group())
	else:
		gStart,_ = _findFirstGroup(expr,fromEnd=1)
		return gStart-1
def _findExpEnd(expr):
	m = re.search(r'^([+-]?\d+(\.\d+)?|\w+)',expr)
	if m:
		return len(m.group())
	else:
		_,gEnd = _findFirstGroup(expr,fromEnd=0)
		return gEnd+1

def _splitExpr(txt,delimiter):
	elements = []
	nextElement = 0
	openPar = 0
	for i,c in enumerate(txt):
		if c == '(':
			openPar += 1
		elif c == ')':
			openPar -= 1
		elif c == delimiter and openPar == 0:
			elements.append(txt[nextElement:i])
			nextElement = i + 1
	elements.append(txt[nextElement:])
	return elements

def _findFirstGroup(txt,fromEnd=0):
	if fromEnd:
		txt = txt[::-1]
		gOpen = ")"
		gClose = "("
	else:
		gOpen = "("
		gClose = ")"

	indexStart = txt.index(gOpen) + 1
	openPar = 1
	for i in range(indexStart,len(txt)):
		if txt[i] == gOpen:
			openPar += 1
		elif txt[i] == gClose:
			openPar -= 1
		if openPar == 0:
			if fromEnd:
				return len(txt)-i, len(txt)-indexStart
			else:
				return indexStart, i
	if fromEnd:
		txt = txt[::-1]
	raise ValueError("Unable to parse equation: '{}'".format(txt))

_MAXPERROW = 20
def _listDeclaration(file, name,items,type=0):
	nritems = len(items)
	if type == 0: # strings
		if nritems == 0:
			file.write('const char *{}[1];\n'.format(name))
		else:
			file.write('const char *{}[{}] = {{\n'.format(name,len(items)))
			i = 0
			iRow = 0
			while(i < len(items)):
				if iRow >= _MAXPERROW:
					file.write(',\n')
					iRow = 0
				if iRow == 0:
					file.write('\t"{}"'.format(items[i]))
				else:
					file.write(',"{}"'.format(items[i]))
				i += 1
				iRow += 1
			file.write('};\n')
	elif type == 1: # int
		if nritems == 0:
			file.write('const int {}[1];\n'.format(name))
		else:
			file.write('const int {}[{}] = {{\n'.format(name,len(items)))
			i = 0
			iRow = 0
			while(i < len(items)):
				if iRow >= _MAXPERROW:
					file.write(',\n')
					iRow = 0
				if iRow == 0:
					file.write('\t{}'.format(items[i]))
				else:
					file.write(',{}'.format(items[i]))
				i += 1
				iRow += 1
			file.write('};\n')
	elif type == 2: # double
		if nritems == 0:
			file.write('const double {}[1];\n'.format(name))
		else:
			file.write('const double {}[{}] = {{\n'.format(name,len(items)))
			i = 0
			iRow = 0
			while(i < len(items)):
				if iRow >= _MAXPERROW:
					file.write(',\n')
					iRow = 0
				if iRow == 0:
					file.write('\t{}'.format(items[i]))
				else:
					file.write(',{}'.format(items[i]))
				i += 1
				iRow += 1
			file.write('};\n')

def _getVariableDependency(model,variables):
	for variable in variables:
		dependency = []
		statedependent = False
		expression = variables[variable]["expression"]
		varinexpr = re.findall(r'\b\w+\b(?!\()',expression)

		for var in varinexpr:
			if var in model["INPUTS"]:
				dependency.append(list(model["INPUTS"]).index(var))
			elif var in variables:
				if list(variables).index(var) >= list(variables).index(variable):
					raise ValueError(f"Variable '{variable}' depends on variable '{var}' which is defined after variable '{variable}'")
				else:
					dependency.extend(variables[var]["inputdependency"])
					if variables[var]["statedependent"]:
						statedependent = True
			elif var in model["STATES"] or var in ["ddt_" + state for state in model["STATES"]]:
				statedependent = True

		variables[variable]["inputdependency"] = list(set(dependency))
		variables[variable]["statedependent"] = statedependent

def _getInputDependency(model):
	dependency = []

	for _,itemdict in model["OUTPUTS"].items():
		expression = itemdict["expression"]
		varinexpr = re.findall(r'\b\w+\b(?!\()',expression)
		for var in varinexpr:
			if var in model["INPUTS"]:
				dependency.append(list(model["INPUTS"]).index(inp))
			elif var in model["VARIABLES"]:
				dependency.extend(model["VARIABLES"][var]["inputdependency"])
	return list(set(dependency))

def _variableDeclaration(file, items):
	if len(items) > 0:
		file.write('\tdouble ')
		i = 0
		iRow = 0
		while(i < len(items)):
			if iRow >= _MAXPERROW:
				file.write(';\n\tdouble ')
				iRow = 0
			if iRow == 0:
				file.write('{}'.format(items[i]))
			else:
				file.write(',{}'.format(items[i]))
			i += 1
			iRow += 1
		file.write(';\n')
def _replaceddt(txt):
	# replace d/dt(state) with ddt_state to make it a c valid name
	return re.sub(r"d/dt\((\w+)\)",r"ddt_\g<1>",txt)

def _convertAbs(txt):
	return re.sub(r"\babs\b",r"fabs",txt)

def _replaceExpr(txt):
	comment = txt.find('//')
	if comment > -1:
		txt = txt[:comment]
	return _convertAbs(_convertPowerExpr(_convertToDouble(_replaceddt(_removeWhiteSpace(txt)))))

def _containsDerivate(state,exprs):
	for expr in exprs:
		if re.search(r'(d/dt\(' + state + r'\)|ddt_' + state + r')', expr):
			return True
	return False

def _removeComments(string):
	string = re.sub(re.compile(r"/\*.*?\*/",re.DOTALL), "", string) # remove all occurrences streamed comments (/*COMMENT */) from string
	string = re.sub(re.compile(r"//[^\n]*" ), "", string) # remove all occurrence single-line comments (//COMMENT\n ) from string
	return string
